/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: ByteImage.c,v 0.9.1.3 1991/12/09 19:35:10 leei Exp $";

#include <math.h>
#include <stdlib.h>
#include <memory.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <vim/ByteImageP.h>
#include <vim/CMManage.h>
#include <vim/gray.h>
#include <vim/gamma.h>

#include <vim/proto.h>
#include <vim/debug.h>

extern double rint __((double));

static XtResource resources[] = {
    { ImNdata, ImCImageData, ImRImageData, sizeof(XtPointer),
         XtOffset(ByteImage,byte.data), XtRImmediate, (XtPointer) 0},
    { ImNdepth, ImCDepth, XtRShort, sizeof(short),
         XtOffset(ByteImage,byte.depth), XtRImmediate, (XtPointer) 8}
};

static void
  Initialize _DCL((im, argv, argc),
		  GenericImage im _AND ArgList argv _AND Cardinal argc );
static void
  Destroy _DCL((im), GenericImage im );

static void
  InitializeParams _DCL((disp), GenericDisplayParams disp );
static Boolean
  SetParams _DCL((old, req, disp), 
		  GenericDisplayParams old _AND 
		  GenericDisplayParams req _AND 
		  GenericDisplayParams disp);
static void
  EditParams _DCL((disp), GenericDisplayParams disp );
static void
  DescribeElem _DCL((im, i, j, buf, len), 
		     GenericImage im _AND int i _AND int j _AND 
		     char *buf _AND int len );

static int
  AllocColors _DCL((disp), GenericDisplayParams disp );
static void
  FreeColors _DCL((disp), GenericDisplayParams disp );
static void
  Redisplay _DCL((disp, same), GenericDisplayParams disp _AND Boolean same );

ByteImageClassRec byteImageClassRec = {
  { /* core fields */
    /* superclass		*/	&genericImageClassRec,
    /* class_name		*/	"ByteImage",
    /* image_size		*/	sizeof(ByteImageRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* resources                */	resources,
    /* num_resources            */	XtNumber(resources),
    /* set_values               */      NULL,
    /* read			*/	NULL,
    /* write			*/	NULL,
    /* image_display_size	*/	sizeof(ByteDisplayParamsRec),
    /* param_resources          */	NULL,
    /* num_param_resources      */	0,
    /* initialize_params	*/	InitializeParams,
    /* set_params		*/	SetParams,
    /* destroy_params		*/	NULL,
    /* edit_params		*/	EditParams,
    /* remember_params		*/	NULL,
    /* compare_params		*/	NULL,
    /* install			*/	NULL,
    /* uninstall		*/	NULL,
    /* alloc_colors             */      AllocColors,
    /* free_colors              */      FreeColors,
    /* display			*/	Redisplay,
    /* description		*/	NULL,
    /* element_description	*/	DescribeElem,
    /* destroy			*/      Destroy
  }
};

ImageClass byteImageClass = (ImageClass) &byteImageClassRec;

static void
  Initialize _IMPL((gim, argv, argc),
		   GenericImage gim _AND ArgList argv _AND Cardinal argc )
{
    ByteImage im = (ByteImage) gim;
    int width  = im->generic.width;
    int height = im->generic.height;
    int depth  = im->byte.depth;

    int numcolors, pixlen, a;

    if ( depth == 0 ) im->byte.depth = depth = 8;

    im->byte.pixlen = pixlen = (depth + 7)/ 8;
    for (numcolors = 2, a = depth - 1; a; a--)
      numcolors *= 2;

    newRGBMapData(&(im->byte.cmap), numcolors);
    if ( !im->byte.data )
      im->byte.data = (byte *) calloc(width * height, pixlen);

    /* No previous pixmap. */
    im->byte.pm = 0;
}

static void
  Destroy _IMPL((gim), GenericImage gim )
{
    ByteImage im = (ByteImage) gim;

    freeRGBMapData(&(im->byte.cmap));
    free(im->byte.data);
}

static void
  RGBCompressRGBMap _IMPL((im), ByteImage im )
{
    int i, j, numbytes;
    byte *dp;
    unsigned short *used_buf;
    use_ByteImage(im,byte);

    used_buf = (unsigned short *) alloca( sizeof(short) * byte_cmap->used );
    for ( i=0; i < byte_cmap->used; ++i ) used_buf[i] = 0;

    /* Mark all CMAP entries used in image. */
    numbytes = byte_width * byte_height;
    for ( i = 0, dp = byte_data; i < numbytes; ++i, ++dp )
      used_buf[*dp] = 1;

    /* Now compress map. */
    for ( i = 0, j = 0; i < byte_cmap->used; ++i )
      if ( used_buf[i] ) {
	  used_buf[i] = j;
	  byte_cmap->red  [j] = byte_cmap->red  [i];
	  byte_cmap->green[j] = byte_cmap->green[i];
	  byte_cmap->blue [j] = byte_cmap->blue [i];
	  ++j;
      }
    byte_cmap->used = j;

    /* And convert image. */
    for ( i = 0, dp = byte_data; i < numbytes; ++i, ++dp )
      *dp = used_buf[*dp];

    time(&(byte_cmap->compressed));
}

static void
  InitializeParams _IMPL((gdisp), GenericDisplayParams gdisp )
{
    ByteDisplayParams disp = (ByteDisplayParams) gdisp;
    ExternalParams *xparams = ImExternalParams(gdisp);

    if ( xparams && DefaultDepthOfScreen(xparams->screen) == 1 )
      chooseGamma(&disp->byte.gamma, gamEnhance);
    else
      defaultGamma(&disp->byte.gamma);
    disp->byte.redo_gamma = True;
}

static void
  DescribeElem _IMPL((gim, i, j, buf, len),
		     GenericImage gim _AND int i _AND int j _AND 
		     char *buf _AND int len )
{
    ByteImage im = (ByteImage) gim;
    ByteImagePart *ipart = &im->byte;
    byte elem = ipart->data[j*im->generic.width + i];
    
    sprintf(buf, "%s: Value %3d { %.2f %.2f %.2f }\n", ImIdent(gim), elem,
	    ipart->cmap.red  [elem]/(float)MaxIntensity,
	    ipart->cmap.green[elem]/(float)MaxIntensity,
	    ipart->cmap.blue [elem]/(float)MaxIntensity);
}

static Boolean
  SetParams _IMPL((g_old, g_req, g_disp), 
		  GenericDisplayParams g_old _AND 
		  GenericDisplayParams g_req _AND 
		  GenericDisplayParams g_disp) 
{
    ByteDisplayParams old  = (ByteDisplayParams) g_old;
    ByteDisplayParams req  = (ByteDisplayParams) g_req;
    ByteDisplayParams disp = (ByteDisplayParams) g_disp;
    
    return disp->byte.redo_gamma;
}

static void
  EditParams _IMPL((gdisp), GenericDisplayParams gdisp )
{
    ByteDisplayParams disp = (ByteDisplayParams) gdisp;
    ExternalParams *xparams = ImExternalParams(gdisp);

    disp->byte.redo_gamma =
      EditGamma(xparams->widget, &disp->byte.gamma);
}

static int
  AllocColors _IMPL((gdisp), GenericDisplayParams gdisp )
{
    ByteDisplayParams disp = (ByteDisplayParams) gdisp;
    ByteImage im = (ByteImage) disp->generic.image;
    ExternalParams *xparams = ImExternalParams(gdisp);
    use_ByteImage(im,byte);
    Visual *vis = xparams->visual;
    GammaParams *gamma = &disp->byte.gamma;
    int i, nalloc;
    Boolean alloc, new_index = False;
    XColor *xcolor, *colors;

    /* Allocate colors (only if a dynamic colormap is involved). */
    if ( byte_cmap->compressed < im->generic.timestamp )
      RGBCompressRGBMap(im);
    alloc = DynamicVisual(vis) &&
      (disp->byte.color_timestamp < im->generic.timestamp ||
       disp->byte.redo_gamma);
    
    if ( alloc && !disp->byte.index ) {
	disp->byte.index = (Pixel *) malloc(sizeof(Pixel) * byte_cmap->used);
	new_index = True;
    }

    if ( !disp->byte.intens )
      disp->byte.intens = (float *) malloc(sizeof(float) * byte_cmap->used);

    colors = (XColor *) alloca(sizeof(XColor) * byte_cmap->used);
    
    /* Initialize the color list. */
    for ( i = 0, xcolor = colors; i < byte_cmap->used; i++, ++xcolor ) {
	xcolor->flags = DoRed | DoGreen | DoBlue;
	xcolor->red   = byte_cmap->red[i];
	xcolor->green = byte_cmap->green[i];
	xcolor->blue  = byte_cmap->blue[i];
	disp->byte.intens[i] = gammaCorrect(gamma, xcolor);
    }

    if ( !alloc ) return 1;

    if ( disp->byte.color_timestamp > 0 )
      CMFreeColors( disp->byte.index, byte_cmap->used );

    /* Now try to get all colors exactly. */
    nalloc = CMAllocColors(colors, byte_cmap->used, /*exact*/True, /*all*/True);

    if ( nalloc != byte_cmap->used ) {
	/* For the moment, this is the cheap way out.  Eventually, 
	 * I'll back off and dither.  Now I just allocate as many as I 
	 * can and then use closest.  ** UPDATE **
	 */
	nalloc = CMAllocColors(colors, byte_cmap->used, False, True);
    }

    /*
     * Now update the pixel entries.
     */
    disp->byte.redo_gamma = False;
    for ( i = 0, xcolor = colors; i < byte_cmap->used; i++, ++xcolor ) {
	if ( !new_index && disp->byte.index[i] != xcolor->pixel )
	  disp->byte.redo_gamma = True;
	disp->byte.index[i] = xcolor->pixel;
    }

    time(&(disp->byte.color_timestamp));
    return 1;
}

static void
  FreeColors _IMPL((gdisp), GenericDisplayParams gdisp )
{
    ByteDisplayParams disp = (ByteDisplayParams) gdisp;
    ByteImage im = (ByteImage) disp->generic.image;
    use_ByteImage(im,byte);

    if ( disp->byte.color_timestamp > 0 )
      CMFreeColors( disp->byte.index, byte_cmap->used );

    if ( disp->byte.index ) {
	free(disp->byte.index);
	disp->byte.index = 0;
    }

    if ( disp->byte.intens ) {
	free(disp->byte.intens);
	disp->byte.intens = 0;
    }

    disp->byte.color_timestamp = 0;
}

static byte *
  pixelateImage _IMPL((im, im_width, im_height, index, 
		       x, y, width, height, depth, llen),
		      byte *im _AND int im_width _AND int im_height _AND 
		      Pixel *index _AND
		      int x _AND int y _AND int width _AND int height _AND 
		      int depth _AND int *llen )
{
    int i, j;
    int bytes = (depth+7)/8;
    byte *dp = &im[im_width*y], *p;
    int numpixels = width * height;

    *llen = bytes * width;
    if ( bytes == sizeof(byte) ) {
	byte *bp, *bdata = (byte *) malloc( sizeof(byte) * numpixels );
	for ( j=0, bp=bdata; j < height; ++j, dp += im_width )
	  for ( i=0, p=dp+x ; i < width; ++i,++p,++bp )
	    *bp = index[*p];
	return bdata;
    } else if ( bytes == sizeof(short) ) {
	short *sp, *sdata = (short *) malloc( sizeof(short) * numpixels );
	for ( j=0, sp=sdata; j < height; ++j, dp += im_width )
	  for ( i=0, p=dp+x ; i < width; ++i,++p,++sp )
	    *sp = index[*p];
	return (byte *) sdata;
    } else if ( bytes == sizeof(int) ) {
	int *ip, *idata = (int *) malloc( sizeof(int) * numpixels );
	for ( j=0, ip=idata; j < height; ++j, dp += im_width )
	  for ( i=0, p=dp+x ; i < width; ++i,++p,++ip )
	    *ip = index[*p];
	return (byte *) idata;
    } else if ( bytes == sizeof(long) ) {
	long *lp, *ldata = (long *) malloc( sizeof(long) * numpixels );
	for ( j=0, lp=ldata; j < height; ++j, dp += im_width )
	  for ( i=0, p=dp+x ; i < width; ++i,++p,++lp )
	    *lp = index[*p];
	return (byte *) ldata;
    } else {
	byte *bp, *bdata = (byte *) malloc( bytes * numpixels );
	for ( j=0, bp=bdata; j < height; ++j, dp += im_width )
	  for ( i=0, p=dp+x ; i < width; ++i,++p, bp += bytes )
	    memcpy(bp,&index[*p],bytes);
	return bdata;
    }
}

static byte *
  zoomByteImage _DCL((im_data, im_width, im_height, cmap, xmag, ymag, 
		      xp, yp, widthp, heightp), 
		     byte *im_data _AND int im_width _AND int im_height _AND 
		     RGBMap *cmap _AND 
		     float xmag _AND float ymag _AND 
		     int *xp _AND int *yp _AND int *widthp _AND int *heightp);

static byte *
  staticGrayDither _DCL((im_data, im_width, im_height, cmap, intens, gamma,
			 x, y, width, height, depth, llen), 
			byte *im_data _AND int im_width _AND int im_height _AND
			RGBMap *cmap _AND float *intens _AND 
			GammaParams *gamma _AND
			int x _AND int y _AND int width _AND int height _AND 
			int depth _AND int *llen);

static byte *
  trueColorDither _DCL((im_data, im_width, im_height, cmap, vis, gamma,
			x, y, width, height, depth, llen), 
		       byte *im_data _AND int im_width _AND int im_height _AND 
		       RGBMap *cmap _AND Visual *vis _AND 
		       GammaParams *gamma _AND
		       int x _AND int y _AND int width _AND int height _AND 
		       int depth _AND int *llen);

void
DisplayByteImage _IMPL((im, im_width, im_height, dst_x, dst_y,
			xparams, cmap, index, intens, gamma,
			pm_return, pm_width, pm_height), 
		       byte *im _AND int im_width _AND int im_height _AND
		       int dst_x _AND int dst_y _AND
		       ExternalParams *xparams _AND 
		       RGBMap *cmap _AND Pixel *index _AND 
		       float *intens _AND GammaParams *gamma _AND
		       Pixmap *pm_return _AND 
		       Dimension *pm_width _AND Dimension *pm_height)
{
    Display *display = xparams->display;
    Window win = xparams->win;
    Visual *vis = xparams->visual;
    Screen *scrn = xparams->screen;
    float mag_x = xparams->mag_x;
    float mag_y = xparams->mag_y;
    Region region = xparams->region;
    XRectangle *range = &xparams->range;
    int ddepth = DefaultDepthOfScreen(scrn);
    int llen;
    int src_x = 0, src_y = 0, zim_width;
    Pixmap pm;
    XImage *xim = 0;
    byte *zim = 0, *zim1 = 0;
    GC gc = 0; XGCValues gcv;
    int width, height, format;
	
    /* Create an image on the given range. */
    width  = range->width;  if ( width  > im_width  ) width  = im_width;
    height = range->height; if ( height > im_height ) height = im_height;
    src_x = range->x; 
    src_y = range->y;

    DEBUG((DBG_DISP, "Display byte image %#x (%dx%d) at %d,%d\n", 
	   im, im_width, im_height, dst_x, dst_y));

    if ( mag_x == 1.0 && mag_y == 1.0 ) {
	zim = im; 
    } else {
	zim = zoomByteImage(im, im_width, im_height, cmap,
			    mag_x, mag_y,
			    &src_x,&src_y, &width,&height);
	if ( !zim ) goto cleanup;
	DEBUG((DBG_DISP, "Zoom image %#x at %d,%d by %g,%g => %#x (%dx%d)\n", 
	       im, src_x, src_y, mag_x, mag_y, zim, width, height));
	im_width = width;
	im_height = height;
	src_x = src_y = 0;
    }
	
    switch ( vis->class ) {
      case StaticColor:
      case TrueColor:
	zim = trueColorDither(zim1=zim, im_width, im_height, cmap, 
			      vis, gamma,
			      src_x,src_y, width,height,ddepth, &llen);
	DEBUG((DBG_DISP, "TC dither image %#x (%dx%d) at %d,%d => %#x\n", 
	       zim1, im_width, im_height, src_x, src_y, zim));
	if ( zim1 != im ) free(zim1);
	zim1 = 0;
	format = ZPixmap;
	src_x = src_y = 0;
	break;
      case StaticGray:
	zim = staticGrayDither(zim1=zim, im_width, im_height, 
			       cmap, intens, gamma,
			       src_x,src_y, width,height,ddepth, &llen);
	DEBUG((DBG_DISP, "SG dither image %#x (%dx%d) at %d,%d %dx%d => %#x\n", 
	       zim1, im_width, im_height, src_x, src_y, width, height, zim));
	if ( zim1 != im ) free(zim1);
	zim1 = 0;
	format = ddepth==1? XYBitmap: ZPixmap;
	src_x = src_y = 0;
	break;
      case GrayScale:
      case PseudoColor:
	zim = pixelateImage(zim1=zim, im_width, im_height, index,
			    src_x,src_y, width,height,ddepth, &llen);
	DEBUG((DBG_DISP, "Pixelate image %#x (%dx%d) at %d,%d => %#x\n", 
	       zim1, im_width, im_height, src_x, src_y, zim));
	if ( zim1 != im ) free(zim1);
	zim1 = 0;
	format = ZPixmap;
	src_x = src_y = 0;
	break;
      default:
	fprintf(stderr,"Cannot handle visual type %d\n",vis->class);
	goto cleanup;
    }

    if ( !zim ) goto cleanup;
    xim = XCreateImage(display, vis, ddepth, format,
		       0, (char*) zim, width, height, 8, llen );
    xim->bitmap_bit_order = MSBFirst;
    xim->byte_order = MSBFirst;
    DEBUG((DBG_DISP, "Create image from %#x (%dx%d) => %#x\n", 
	   zim, width, height, xim));

    /* Create the appropriate GC. */
    gcv.function = GXcopy;
    gcv.foreground = CMwhitePixel;
    gcv.background = CMblackPixel;
    gc = XCreateGC(display, win, GCFunction|GCForeground|GCBackground, &gcv);
    
    if ( xim ) {
	/* Finally draw it. */
	pm = XCreatePixmap(display, win, width, height, ddepth);
	if ( pm_return ) *pm_return = pm;
	if ( pm_width )  *pm_width = width;
	if ( pm_height ) *pm_height = height;
	DEBUG((DBG_DISP, "Put image %#x at %d,%d (%dx%d) into pixmap %#x\n", 
	       xim, src_x, src_y, width, height, pm));
	XPutImage(display, pm, gc, xim, src_x, src_y, 0, 0, width, height);
    }

    XSetRegion( display, gc, region ); /**/
    DEBUG((DBG_DISP, "CopyArea %#x (%dx%d) to %d,%d\n", 
	   pm, width, height, dst_x, dst_y));
    XCopyArea(display, pm, win, gc, 0, 0, width, height, dst_x, dst_y);
    
    /* And clean up. */
cleanup:
    if ( zim && zim != im ) free(zim);
    if ( xim ) free(xim);
    if ( gc )  XFreeGC(display,gc);
    if ( pm && !pm_return ) XFreePixmap(display, pm);
}

		  
static void
  Redisplay _IMPL((gdisp, same), 
		  GenericDisplayParams gdisp _AND Boolean same )
{
    ByteDisplayParams disp = (ByteDisplayParams) gdisp;
    ByteImage im = (ByteImage) disp->generic.image;
    ExternalParams *xparams = ImExternalParams(gdisp);
    Display *display = xparams->display;
    Window win = xparams->win;
    Visual *vis = xparams->visual;
    Region region = xparams->region;
    float mag_x = xparams->mag_x;
    float mag_y = xparams->mag_y;
    XRectangle *range = &xparams->range;
    int dst_x, dst_y;
    Pixmap pm = im->byte.pm;
    GC gc = 0; XGCValues gcv;
    use_ByteImage(im,byte);
    
    dst_x = (int) rint(mag_x*range->x);
    dst_y = (int) rint(mag_y*range->y);

    DEBUG((DBG_DISP, "ByteImage::Redisplay(%#x, %d) redo_gamma = %d\n",
	   disp, same, disp->byte.redo_gamma));

    if ( !same || !pm || disp->byte.redo_gamma ) {
	int format;
	
	/* Free the previous pixmap. */
	if ( pm ) XFreePixmap(display,pm);
	
	DisplayByteImage(byte_data, byte_width, byte_height,
			 dst_x, dst_y,
			 xparams, byte_cmap, disp->byte.index,
			 disp->byte.intens, &disp->byte.gamma,
			 &im->byte.pm, 
			 &im->byte.pm_width, &im->byte.pm_height);
	disp->byte.redo_gamma = False;
    } else {
	/* Create the appropriate GC. */
	gcv.function = GXcopy;
	gcv.foreground = CMwhitePixel;
	gcv.background = CMblackPixel;
	gc = XCreateGC(display, win, 
		       GCFunction|GCForeground|GCBackground, &gcv);
	XSetRegion( display, gc, region ); /**/
	XCopyArea(display, pm, win, gc, 
		  0, 0, im->byte.pm_width, im->byte.pm_height, 
		  dst_x, dst_y);
    }
}

static unsigned int *
  buildIndex _IMPL((x, width, mag, rx, rwidth),
		   int x _AND int width _AND float mag _AND 
		   int *rx _AND int *rwidth)
{
    unsigned int *index;
    unsigned int  a;

    *rwidth = (int) rint(mag * width);
    *rx = 0;

    index = (unsigned int *) malloc(*rwidth * sizeof(unsigned int));

    for (a = *rx; a < *rwidth; a++)
      index[a] = a/mag + x ;
    return(index);
}

static int abs _IMPL((x), int x) { return x>0? x: -x; }

static int
  RGBdist _IMPL((r0, g0, b0, r1, g1, b1),
		Intensity r0 _AND Intensity g0 _AND Intensity b0 _AND 
		Intensity r1 _AND Intensity g1 _AND Intensity b1)
{
    return abs(r0-r1) + abs(g0-g1) + abs(b0-b1);
}

static int
  BestInMap _IMPL((map, r, g, b), 
		  RGBMap *map _AND 
		  Intensity r _AND Intensity g _AND Intensity b )
{
    int i, used = map->used;
    Intensity *rp = map->red;
    Intensity *gp = map->green;
    Intensity *bp = map->blue;
    short best = 0;
    int dist = abs(r-*rp++) + abs(g-*gp++) + abs(b-*bp++);
    for ( i = 1; i < used; ++i ) {
	int nd = abs(r-*rp++) + abs(g-*gp++) + abs(b-*bp++);
	if ( nd < dist ) { best = i; dist = nd; }
	if ( nd == 0 ) break;
    }
    return best;
}

static byte *
  zoomByteImage _IMPL((im_data, im_width, im_height, cmap, xmag, ymag, 
		       xp, yp, widthp, heightp), 
		      byte *im_data _AND int im_width _AND int im_height _AND 
		      RGBMap *cmap _AND 
		      float xmag _AND float ymag _AND 
		      int *xp _AND int *yp _AND int *widthp _AND int *heightp)
{
    unsigned int *xindex, *yindex;
    unsigned int  x, y, xsrc, ysrc, pxsrc, pysrc;
    unsigned int  srclinelen;
    unsigned int  destlinelen;
    byte         *dest;
    byte         *srcline, *psrcline, *srcptr, *psrcptr;
    byte         *destline, *destptr;
    unsigned int  srcwidth  = *widthp;
    unsigned int  srcheight = *heightp;
    unsigned int  srcx = *xp;
    unsigned int  srcy = *yp;
    byte src;

    xindex = buildIndex(srcx, srcwidth,  xmag, xp, widthp);
    yindex = buildIndex(srcy, srcheight, ymag, yp, heightp);

    destlinelen = *widthp;
    dest = destline = (byte *) calloc( *heightp * destlinelen, 1 );

    srclinelen = im_width;
    srcline = im_data + srcy * srclinelen + srcx;
    
    if ( xmag < 1.0 || ymag < 1.0 ) {
	for (y = 0, ysrc = yindex[y]; y < *heightp; y++) {
	    pysrc    = ysrc;
	    psrcline = srcline;
	    
	    ysrc     = yindex[y];
	    srcline += (ysrc-pysrc) * srclinelen;
	    
	    srcptr   = srcline;
	    destptr  = destline;
	    
	    for (x = 0, xsrc = xindex[x]; x < *widthp; x++) {
		short n = 0;
		unsigned short r, g, b, pr, pg, pb;
		int srcr = 0, srcg = 0, srcb = 0;
		short pxl = -1;

		pxsrc    = xsrc;
		xsrc     = xindex[x];
		srcptr  += (xsrc-pxsrc);
		psrcptr  = srcptr;
		psrcline = srcline;

		do {
		    do {
			++pxsrc; ++n;
			srcr += r = cmap->red  [*psrcptr]; 
			srcg += g = cmap->green[*psrcptr];
			srcb += b = cmap->blue [*psrcptr];
			if ( pxl < 0 || 
			     RGBdist(r,g,b, srcr/n,srcg/n,srcb/n) <
			     RGBdist(pr,pg,pb, srcr/n,srcg/n,srcb/n) )
			  pxl = *psrcptr, pr = r, pg = g, pb = b;
			--psrcptr;
		    } while ( xsrc > pxsrc );
		    ++pysrc;
		    psrcline -= srclinelen;
		    psrcptr   = psrcline + xsrc;
		} while ( ysrc > pysrc );

		*destptr++ = pxl;
	    }
	    destline += destlinelen;
	}
    } else {
	for (y = 0, ysrc = yindex[y]; y < *heightp; y++) {
	    pysrc    = ysrc;
	    psrcline = srcline;
	    
	    ysrc     = yindex[y];
	    if ( ysrc > pysrc )
	      srcline += (ysrc-pysrc) * srclinelen;
	    
	    srcptr   = srcline;
	    destptr  = destline;
	    
	    for (x = 0, xsrc = xindex[x]; x < *widthp; x++) {
		pxsrc   = xsrc;
		xsrc    = xindex[x];
		if ( xsrc > pxsrc )
		  srcptr += (xsrc-pxsrc);
		
		src = *srcptr;
		*destptr++ = src;
	    }
	    destline += destlinelen;
	}
    }

    free((char *)xindex);
    free((char *)yindex);

    return dest;
}

static void 
  GLeftToRight _DCL((inc, curr, next, width),
		    short inc _AND short *curr _AND short *next _AND 
		    short width);
static void 
  GRightToLeft _DCL((inc, curr, next, width),
		    short inc _AND short *curr _AND short *next _AND 
		    short width);

static byte *
  staticGrayDither _IMPL((im, im_width, im_height, cmap, intens, gamma,
			  x, y, width, height, depth, llen), 
			 byte *im _AND 
			 int im_width _AND int im_height _AND
			 RGBMap *cmap _AND float *intens _AND 
			 GammaParams *gamma _AND
			 int x _AND int y _AND int width _AND int height _AND 
			 int depth _AND int *llen)
{
    int used = cmap->used;
    byte perpxl = 8/depth;
    short destshft;
    unsigned short *gray = 0;
    short destlinelen;
    short maxval = (1<<depth)-1;
    short incval = (MaxGrey+maxval)/maxval;
    byte *dest, *destptr, *destline;
    byte *srcline;
    short *curr, *next, *swap;
    int i, j;

    if ( depth > 8 ) {
	fprintf(stderr, "Unable to dither for %d bit displays.\n", depth );
	return 0;
    }

    if ( !intens ) {
	/* Initialize the color list. */
	intens = (float *) alloca(sizeof(float) * used);
	for ( i = 0; i < used; i++ ) {
	    XColor xcolor;
	    xcolor.flags = DoRed | DoGreen | DoBlue;
	    xcolor.red   = cmap->red[i];
	    xcolor.green = cmap->green[i];
	    xcolor.blue  = cmap->blue[i];
	    intens[i] = gammaCorrect(gamma, &xcolor);
	}
    }

    /* Setup gray scaling array. */
    if ( depth <= 16 ) {
	gray = (unsigned short *) alloca(sizeof(unsigned short) * used);
	for ( i=0; i<used; ++i ) {
	    gray[i] = grayIntens(intens[i]);
	}
    }

    /* Setup circulating line buffers. */
    curr = (short *) alloca( sizeof(short) * (width+2) ); ++curr;
    next = (short *) alloca( sizeof(short) * (width+2) ); ++next;
    for ( i=0; i<width+2; ++i ) curr[i] = next[i] = 0;

    srcline = im + y*im_width + x;

    *llen = destlinelen = (width+perpxl-1)/perpxl;
    dest = destline = (byte *) calloc( height * destlinelen, 1 );
	
    for (j = 0; j < height; j++ ) {
	destptr  = destline;
	destshft = (8-depth);
	  
	for (i = 0; i < width; i++ ) {
	    byte src = srcline[i];
	    curr[i+1] += (gray)? gray[src]: grayIntens(intens[i]);
	}

	/* Either FS dither from left->right or right->left (alternate). */
	if ( j & 1 )
	  GLeftToRight( incval, curr+1, next+1, width );
	else
	  GRightToLeft( incval, curr+1, next+1, width );
	  
	/* Pack the FS dithered values into the pixmap */
	if ( depth == 8 ) {
	    for (i = 0; i < width; i++ )
	      *destptr++ = curr[i+1]/incval;
	} else {
	    for (i = 0; i < width; i++ ) {
		unsigned short val = curr[i+1]/incval;
		if (val) *destptr |= (val<<destshft);
		if ( (destshft -= depth) < 0) {
		    destshft = (8-depth);
		    destptr++;
		}
	    }
	}
	srcline  += im_width;
	destline += destlinelen;

	/* swap curr and next */
	swap = curr; curr = next; next = swap;
	for ( i=0; i<width+2; ++i ) next[i] = 0;
    }

    return dest;
}


/*
 * dither a line from left to right
 */
static void 
  GLeftToRight _IMPL((inc, curr, next, width),
		     short inc _AND short *curr _AND short *next _AND 
		     short width)
{
    short error;
    short idx;
    unsigned short output;
    unsigned short off = inc/2;

    for (idx=0; idx<width; idx++) {
	output       = (curr[idx]+off)/inc;
	output      *= inc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx-1] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx+1] += error * 1 / 16;
	curr[idx+1] += error * 7 / 16;
    }
}


/*
 * dither a line from right to left
 */
static void 
  GRightToLeft _IMPL((inc, curr, next, width),
		     short inc _AND short *curr _AND short *next _AND 
		     short width)
{
    short error;
    short idx;
    unsigned short output;
    unsigned short off = inc/2;

    for (idx=(width-1); idx>=0; idx--) {
	output       = (curr[idx]+off)/inc;
	output      *= inc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx+1] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx-1] += error * 1 / 16;
	curr[idx-1] += error * 7 / 16;
    }
}

static void 
  CLeftToRight _DCL((rinc, ginc, binc, curr, next, width),
		     short rinc _AND short ginc _AND short binc _AND 
		     short *curr _AND short *next _AND short width);
static void 
  CRightToLeft _DCL((rinc, ginc, binc, curr, next, width),
		     short rinc _AND short ginc _AND short binc _AND 
		     short *curr _AND short *next _AND short width);

static void
  decode_mask _IMPL((mask, max, inc, shft),
		    int mask _AND short *max _AND short *inc _AND short *shft)
{
    *shft = 0; *max = 1;
    while ( !(mask & 0x01) ) {
	mask >>= 1;
	++*shft;
    }
    while ( mask & 0x01 ) {
	mask >>= 1;
	*max <<= 1;
    }
    --*max;
    *inc = (MaxGrey + *max) / *max;
}

static byte *
  trueColorDither _IMPL((im, im_width, im_height, cmap, vis, gamma,
			 x, y, width, height, depth, llen), 
			byte *im _AND 
			int im_width _AND int im_height _AND 
			RGBMap *cmap _AND Visual *vis _AND 
			GammaParams *gamma _AND
			int x _AND int y _AND int width _AND int height _AND 
			int depth _AND int *llen)
{
    byte perpxl = 8/depth;
    short destshft;
    short destlinelen;
    short rmax, rinc, rshft;
    short gmax, ginc, gshft;
    short bmax, binc, bshft;
    byte *dest, *destptr, *destline;
    byte *srcline, *sp;
    short *curr, *next, *swap;
    int i, j;

    decode_mask( vis->red_mask,   &rmax, &rinc, &rshft );
    decode_mask( vis->green_mask, &gmax, &ginc, &gshft );
    decode_mask( vis->blue_mask,  &bmax, &binc, &bshft );
    
#define ENCODE(r,g,b)\
    (((r/rinc)<<rshft) + ((g/ginc)<<gshft) + ((b/binc)<<bshft))

    if ( depth > 8 ) {
	fprintf(stderr, "Unable to dither for %d bit displays.\n", depth );
	return 0;
    }

    /* Setup circulating line buffers. */
    curr = (short *) alloca( 3 * sizeof(short) * (width+2) ); ++curr;
    next = (short *) alloca( 3 * sizeof(short) * (width+2) ); ++next;
    for ( i=0; i<3*(width+2); ++i ) curr[i] = next[i] = 0;

    srcline = im + y*im_width + x;

    *llen = destlinelen = (width+perpxl-1)/perpxl;
    dest = destline = (byte *) calloc( height * destlinelen, 1 );
	
    for (j = 0; j < height; j++ ) {
	destptr  = destline;
	destshft = (8-depth);
	
	for (i = 3, sp = srcline; i < 3*(width+1); i += 3, ++sp ) {
	    XColor xcolor;
	    xcolor.red   = cmap->red[*sp];
	    xcolor.green = cmap->green[*sp];
	    xcolor.blue  = cmap->blue[*sp];
	    gammaCorrect(gamma, &xcolor);
	    curr[i+0] += intensGray(xcolor.red);
	    curr[i+1] += intensGray(xcolor.green);
	    curr[i+2] += intensGray(xcolor.blue);
	}

	/* Either FS dither from left->right or right->left (alternate). */
	if ( j & 1 )
	  CRightToLeft( rinc,ginc,binc, curr+3, next+3, width );
	else
	  CLeftToRight( rinc,ginc,binc, curr+3, next+3, width );
	
	/* Pack the FS dithered values into the pixmap */
	if ( depth == 8 ) {
	    for (i = 3; i < 3*(width+1); i += 3 )
	      *destptr++ = ENCODE(curr[i], curr[i+1], curr[i+2]);
	} else {
	    for (i = 3; i < 3*(width+1); i += 3 ) {
		byte val = ENCODE(curr[i], curr[i+1], curr[i+2]);
		if (val) *destptr |= (val<<destshft);
		if ( (destshft -= depth) < 0) {
		    destshft = (8-depth);
		    destptr++;
		}
	    }
	}
	srcline  += im_width;
	destline += destlinelen;

	/* swap curr and next */
	swap = curr; curr = next; next = swap;
	for ( i=0; i<3*(width+2); ++i ) next[i] = 0;
    }
#undef ENCODE

    return dest;
}


/*
 * dither a line from left to right
 */
static void 
  CLeftToRight _IMPL((rinc, ginc, binc, curr, next, width),
		     short rinc _AND short ginc _AND short binc _AND 
		     short *curr _AND short *next _AND short width)
{
    short error;
    short idx;
    unsigned short output;
    unsigned short roff = rinc/2;
    unsigned short goff = ginc/2;
    unsigned short boff = binc/2;

    for ( idx = 0; idx < 3*width; ) {
	/* Red */
	output       = (curr[idx]+roff)/rinc;
	output      *= rinc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx-3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx+3] += error * 1 / 16;
	curr[idx+3] += error * 7 / 16;
	++idx;

	/* Green */
	output       = (curr[idx]+goff)/ginc;
	output      *= ginc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx-3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx+3] += error * 1 / 16;
	curr[idx+3] += error * 7 / 16;
	++idx;
	
	/* Blue */
	output       = (curr[idx]+boff)/binc;
	output      *= binc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx-3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx+3] += error * 1 / 16;
	curr[idx+3] += error * 7 / 16;
	++idx;
    }
}


/*
 * dither a line from right to left
 */
static void 
  CRightToLeft _IMPL((rinc, ginc, binc, curr, next, width),
		     short rinc _AND short ginc _AND short binc _AND 
		     short *curr _AND short *next _AND short width)
{
    short error;
    short idx;
    unsigned short output;
    unsigned short roff = rinc/2;
    unsigned short goff = ginc/2;
    unsigned short boff = binc/2;

    for ( idx = (3*width-1); idx >= 0; ) {
	/* Blue */
	output       = (curr[idx]+boff)/binc;
	output      *= binc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx+3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx-3] += error * 1 / 16;
	curr[idx-3] += error * 7 / 16;
	--idx;

	/* Green */
	output       = (curr[idx]+goff)/ginc;
	output      *= ginc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx+3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx-3] += error * 1 / 16;
	curr[idx-3] += error * 7 / 16;
	--idx;

	/* Red */
	output       = (curr[idx]+roff)/rinc;
	output      *= rinc;
	error        = curr[idx] - output;
	curr[idx]    = output;
	next[idx+3] += error * 3 / 16;
	next[idx]   += error * 5 / 16;
	next[idx-3] += error * 1 / 16;
	curr[idx-3] += error * 7 / 16;
	--idx;
    }
}


