/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: CMManage.c,v 0.9.1.3 1991/12/09 19:35:26 leei Exp $";

/* 
 * Contains these definitions:
 *    CMInitialize()
 *    CMAllocColor()
 *    CMAllocNamedColor()
 *    CMAllocColorCells()
 *    CMAllocColorPlanes()
 *    CMFreeColors()
 * Temporary:
 *    CMCreateColormap()
 *    CMCopyColormapAndFree()
 *    CMFreeColormap()
 */

#if defined(__GNUC__) && !defined(alloca)
#define alloca __builtin_alloca
#elif defined(sun)
#include <alloca.h>
#else
extern char *alloca();
#endif

#if !defined(__GNUC__) 
#define inline
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xmu/CharSet.h>

#include <vim/CMManage.h>
#include <vim/debug.h>
#include <vim/proto.h>
#include <vim/utils.h>

Display      *CMdisplay;
Screen       *CMscreen;
Visual       *CMvisual;
Colormap      CMcmap = 0;
unsigned int  CMdepth;
Pixel         CMblackPixel;
Pixel         CMwhitePixel;

static Widget   CMshell = 0;
static Colormap CMcmap_orig = 0;
static XColor   CMblack;
static XColor   CMwhite;
static Boolean  CMdynamic = 0;

#define CMmap_SIZE (256)

static struct {
    int max_length;
    int length;
    int ncolors;
    int free;
    short *refs;
    XColor *color;
} CMmap;

static void InitializeConverters();

/* 
 * These internal functions are defined and used to manage the color map.
 */

static void
  CMmap_extend _DCL_VOID;
static int
  colorDistance _DCL((c1, c2), XColor *c1 _AND XColor *c2 );
static XColor *
  CMmap_find _DCL((color, exact), XColor *color _AND int exact );
static XColor *
  CMmap_find_pixel _DCL((pixel), Pixel pixel );
static int
  CMmap_alloc _DCL((color, exact), XColor *color _AND int exact );
static void
  CMmap_free_idx _DCL((idx, xfree), int idx _AND int xfree );
static void
  CMmap_reinitialize _DCL_VOID;

static Visual *
  bestVisualOfClassAndDepth _DCL((disp, screen, class, depth),
				 Display *disp _AND Screen *screen _AND 
				 int class _AND unsigned int depth);
static void
  bestVisualOfClass _DCL((disp, screen, visual_class, rvisual, rdepth),
			 Display *disp _AND Screen *screen _AND 
			 int visual_class _AND 
			 Visual **rvisual _AND unsigned int *rdepth);

/* 
 * void 
 * CMInitialize( Widget shell, int class )
 *    - Interface.  Must be called before any other CM functions.  The 
 *      <shell> argument should be the top-level shell widget of the 
 *      application.
 */
void
CMInitialize _IMPL((shell, class), Widget shell _AND int class )
{
    XColor color;

    CMshell = shell;
    CMdisplay = XtDisplay(CMshell);
    CMscreen = XtScreen(CMshell);

    if ( class == -1 ) {
	CMvisual = DefaultVisualOfScreen(CMscreen);
	CMdepth  = DefaultDepthOfScreen(CMscreen);
    } else
      bestVisualOfClass(CMdisplay, CMscreen, class, &CMvisual, &CMdepth);

    CMdynamic = (CMvisual->class & 1);

    if ( CMvisual == DefaultVisualOfScreen(CMscreen) ) {
	XtVaGetValues(shell,XtNcolormap,&CMcmap_orig,NULL);
	CMcmap = CMcmap_orig;
    } else {
	CMcmap = XCreateColormap(CMdisplay, RootWindowOfScreen(CMscreen),
				 CMvisual, AllocNone);
	XtVaSetValues(CMshell, XtNvisual, CMvisual, XtNcolormap, CMcmap, NULL);
    }

    InitializeConverters();

    DEBUG((DBG_CM,"Initial colormap %x\n",CMcmap));


    if ( CMdynamic ) {
	/* Initialize the managed color map. */
	CMmap.max_length = 1<<DefaultDepthOfScreen(XtScreen(shell));
	CMmap.length = MIN(CMmap.max_length,CMmap_SIZE);
	CMmap.ncolors = CMmap.free = 0;
	CMmap.refs = (short *) malloc( sizeof(short) * CMmap.length );
	CMmap.color = (XColor *) malloc( sizeof(XColor) * CMmap.length );

	/* Test QueryColors. */
	color.pixel = CMblackPixel = BlackPixelOfScreen(CMscreen);
	color.flags = DoRed | DoGreen | DoBlue;
	XQueryColor(CMdisplay,CMcmap,&color);
	CMblack = color;
	CMmap_alloc(&color,True);

	color.pixel = CMwhitePixel = WhitePixelOfScreen(CMscreen);
	color.flags = DoRed | DoGreen | DoBlue;
	XQueryColor(CMdisplay,CMcmap,&color);
	CMwhite = color;
	CMmap_alloc(&color,True);
    } else {
	switch ( CMvisual->class ) {
	  case StaticGray:
	  case StaticColor:
	  case TrueColor:
	    CMblackPixel = 0;
	    CMwhitePixel = (1<<CMvisual->bits_per_rgb) - 1;
	    break;
	}
	color.pixel = CMblackPixel;
	color.flags = DoRed | DoGreen | DoBlue;
	XQueryColor(CMdisplay,CMcmap,&color);
	CMblack = color;

	color.pixel = CMwhitePixel;
	color.flags = DoRed | DoGreen | DoBlue;
	XQueryColor(CMdisplay,CMcmap,&color);
	CMwhite = color;
    }

    /* Ensure that we're properly contrasted. */
    if ( CMwhite.red + CMwhite.green + CMwhite.blue <
	 CMblack.red + CMblack.green + CMblack.blue ) {
	color = CMwhite;
	CMwhite = CMblack;
	CMblack = color;
	CMwhitePixel = CMwhite.pixel;
	CMblackPixel = CMblack.pixel;
    }
}

/* 
 * static void
 * CMmap_extend()
 *    - Internal function.  Double the size of the managed map if 
 *      possible.  The map will not grow larger than is supported by 
 *      the bit depth of the display.
 */
static void
CMmap_extend _IMPL_VOID
{
    if ( CMmap.length != CMmap.max_length ) {
	CMmap.length = MIN(CMmap.max_length,2*CMmap.length);
	CMmap.refs = (short *)  
	  realloc( CMmap.refs, sizeof(short)  * CMmap.length );
	CMmap.color = (XColor *) 
	  realloc( CMmap.color, sizeof(XColor) * CMmap.length );
    }
}

/* 
 * static int
 * colorDistance( XColor *c1, XColor *c2 )
 *    - Internal function.  Computes the distance in RGB space between 
 *      <c1> and <c2>.
 */
static int
colorDistance _IMPL((c1, c2), XColor *c1 _AND XColor *c2 )
{
    return 
      ABS((int)c1->red   - (int)c2->red) + 
      ABS((int)c1->green - (int)c2->green) +
      ABS((int)c1->blue  - (int)c2->blue);
}

/*
 * static XColor *
 * CMmap_find( XColor *color, int exact )
 *    - Internal function.  Attempts to find the color defined by 
 *      <color> in the current managed map.  If <exact> is True, then 
 *      only exact RGB matches are accepted, otherwise returns the 
 *      closest.  The pointer returned is a reference into the 
 *      CMmap.color array.
 */
static XColor *
CMmap_find _IMPL((color, exact), XColor *color _AND int exact )
{
    XColor *real = 0, *curr;
    int dist = 4*0xffff; /* Greater than any distance. */
    int idx;
    
    /* All else has failed. Find best already there. */
    for ( idx=0, curr=CMmap.color; idx < CMmap.ncolors; ++idx, ++curr ) {
	if ( CMmap.refs[idx]>=0 ) {
	    if ( exact ) {
		if ( curr->red == color->red && 
		     curr->green == color->green && 
		     curr->blue == color->blue )
		  return curr;
	    } else {
		int new_dist = colorDistance(color,curr);
		if ( new_dist < dist ) {
		    dist = new_dist;
		    real = curr;
		}
	    }
	}
    }
    !exact &&
      (real->red != color->red ||
       real->green != color->green ||
       real->blue != color->blue) &&
      DEBUG((DBG_CM,"%05u: Using %04x/%04x/%04x for %04x/%04x/%04x (D %d)\n",
	     real->pixel, real->red, real->green, real->blue,
	     color->red, color->green, color->blue,
	     colorDistance(real, color)));
    return exact? 0: real;
}

/*
 * static int
 * CMmap_find_pixel( Pixel pixel )
 *    - Internal function.  Attempts to find the color accessed by 
 *      <pixel>.  Returns a color reference from colors or NULL.
 */
static XColor *
CMmap_find_pixel _IMPL((pixel), Pixel pixel )
{
    XColor *curr;
    int idx;
    
    /* All else has failed. Find best already there. */
    for ( idx=0, curr=CMmap.color; idx < CMmap.ncolors; ++idx, ++curr ) {
	if ( CMmap.refs[idx]>=0 && curr->pixel == pixel )
	  return curr;
    }

    return 0;
}

/* 
 * static int
 * CMmap_alloc( XColor *color, int exact )
 *    - Internal function.  Allocate a managed colormap entry with the 
 *      given <color> values.  On return, the pixel entry will contain 
 *      the pixel reference for the given color.  If <exact> is True, 
 *      then only exact color matches are accepted, otherwise the 
 *      nearest color already in the map will be used if all else fails.
 */

static inline int
ColorIsBlack _IMPL((color), XColor *color )
{
    return 
      (color->red   == CMblack.red) && 
      (color->green == CMblack.green) && 
      (color->blue  == CMblack.blue);
}

static inline int
ColorIsWhite _IMPL((color), XColor *color )
{
    return 
      (color->red   == CMwhite.red) && 
      (color->green == CMwhite.green) && 
      (color->blue  == CMwhite.blue);
}

static int
CMmap_alloc _IMPL((color, exact), XColor *color _AND int exact )
{
    XColor *real;
    XColor temp;
    int idx;

    if ( CMdynamic ) {
	if ( real = CMmap_find(color,True) ) {
	    /* Found color.  Use it. */
	    *color = *real;
	    ++CMmap.refs[real-CMmap.color];
	    return real-CMmap.color;
	} else if ( CMmap.free || CMmap.ncolors < CMmap.max_length ) {
	    /* Room to try to allocate another color. */
	    if ( CMmap.free ) {
		/* Reuse a freed slot. */
		idx = CMmap.free;
		CMmap.free = -CMmap.refs[idx]-1;
	    } else {
		/* Extend map and try again. */
		if ( CMmap.ncolors >= CMmap.length ) CMmap_extend();
		idx = CMmap.ncolors++;
	    }
	    /* Now actually allocate the color. */
	    real = &CMmap.color[idx];
	    CMmap.refs[idx] = 1;
	    if ( XAllocColor(CMdisplay,CMcmap,color) ) {
		/* Don't accept reallocations of the black/white pixels. */
		if ( (color->pixel == BlackPixelOfScreen(CMscreen) &&
		      !ColorIsBlack(color)) ||
		     (color->pixel == WhitePixelOfScreen(CMscreen) &&
		      !ColorIsWhite(color)) ) {
		    *real = *color;
		    CMmap_free_idx(idx,True);
		    if ( color->pixel == BlackPixelOfScreen(CMscreen) )
		      temp.red = temp.green = temp.blue = 0;
		    else
		      temp.red = temp.green = temp.blue = 65535;
		    XAllocColor(CMdisplay,CMcmap,&temp);
		    return CMmap_alloc(color,exact);
		}

		/* Real X allocation worked, so just return. */
		DEBUG((DBG_CM, "%05u: Alloc color %4x/%4x/%4x = %d\n", 
		       color->pixel,
		       color->red,color->green,color->blue,idx));
		*real = *color;
		return idx;
	    } else if ( CMcmap == CMcmap_orig ) {
		/* Shared colormap is full, so try a new one. */
		CMmap_free_idx(idx,False);
		CMmap_reinitialize();
		return CMmap_alloc(color,exact);
	    }
	}
	
	if ( !exact ) {
	    /* All else has failed. Find best already there. */
	    real = CMmap_find(color,False);
	    *color = *real;
	    ++CMmap.refs[real-CMmap.color];
	    return real-CMmap.color;
	}
    } else {
	return XAllocColor(CMdisplay, CMcmap, color);
    }
    
    return -1;
}

/* 
 * static void
 * CMmap_free_idx( int idx, int xfree )
 *    - Internal function.  Free the managed map entry with the given 
 *      index.  If this is the last reference to that entry, then the 
 *      associated color in the colormap is freed and the pixel value 
 *      marked for reuse.
 */
static void
CMmap_free_idx _IMPL((idx, xfree), int idx _AND int xfree )
{
    XColor *color = &CMmap.color[idx];
    if ( --CMmap.refs[idx] <= 0 ) {
	if ( xfree ) {
	    DEBUG((DBG_CM,"%05u: Free color %d\n",CMmap.color[idx].pixel,idx));
	    XFreeColors(CMdisplay,CMcmap,&(CMmap.color[idx].pixel),1,0);
	}
	CMmap.refs[idx] = -CMmap.free-1;
	CMmap.free = idx;
    }
}

/* 
 * static void
 * CMmap_reinitialize()
 *    - Internal function.  Create a new colormap and move the 
 *      currently valid entries from the old managed map to the new one.
 */
static void
CMmap_reinitialize _IMPL_VOID
{
    /* Make the new one. */
    CMcmap = XCopyColormapAndFree(CMdisplay, CMcmap);
    DEBUG((DBG_CM,"Copy Colormap and Free\n"));
    XtVaSetValues( CMshell, XtNcolormap, CMcmap, NULL );
}

/* 
 * Status
 * CMAllocColor( XColor *color, int exact )
 *    - Interface.  Allocate a color entry for <color>.  If <exact> is 
 *      True, then only exact color matches will return successfully.
 */
Status
CMAllocColor _IMPL((color, exact), XColor *color _AND int exact )
{
    return (CMmap_alloc( color, exact ) >= 0);
}

/* 
 * int
 * CMAllocColors( XColor *colors, int ncolors, int exact, int all )
 *    - Interface.  Allocate a set of <ncolors> colors as described in 
 *      <colors>.  If <exact> is true then only exact color matches 
 *      will be accepted.  If <all> is true, then no colors are 
 *      allocated unless all of them can be.  Returns the number of 
 *      colors successfully allocated (although if <all> is true, then 
 *      this is the number which `may' be allocated).
 */
int
CMAllocColors _IMPL((colors, ncolors, exact, all),
		    XColor *colors _AND int ncolors _AND 
		    int exact _AND int all )
{
    int i, idx;
    XColor *curr;
    int *index;
    
    DEBUG((DBG_CM,"Attempting to allocate %d colors%s%s.\n",
	   ncolors,exact?" [Exact]":"",all?" [All]":""));
    index = (int *) alloca( sizeof(int) * ncolors );
    for ( idx = 0, curr = colors; idx < ncolors; ++idx, ++curr ) {
	index[idx] = CMmap_alloc(curr,exact);
	if ( index[idx] < 0 ) break;
    }

    /* If all then any lack of success is unacceptable.  Free successful. */
    if ( all && idx < ncolors ) {
	DEBUG((DBG_CM,"Cannot alloc all: Freeing %d colors\n",idx));
	for ( i = 0; i < idx; ++i ) CMmap_free_idx(index[i],True);
    }

    return idx;
}

/* 
 * Status
 * CMAllocNamedColor( char *name, XColor *screen_def, XColor *exact_def )
 *    - Interface.  Allocate the color named <name>.  The <screen_def> 
 *      is filled in with closest valid color supported by the visual, 
 *      while <exact_def> is the exact database definition.  The color 
 *      in <screen_def> will contain the pixel value to be associated.
 */
Status
CMAllocNamedColor _IMPL((name, screen_def, exact_def),
			char *name _AND 
			XColor *screen_def _AND XColor *exact_def )
{
    Status ret = XParseColor(CMdisplay, CMcmap, name, exact_def);
    if ( ret ) {
	*screen_def = *exact_def;
	return CMmap_alloc(screen_def,False);
    } else {
	String params[1];
	Cardinal num_params = 1;
	params[0] = name;
	XtWarningMsg("CMAllocNamedColor", "badFormat", XtCCMError,
		     "RGB color specification \"%s\" has invalid format",
		     params, &num_params);
	return False;
    }
}

void
CMFreeColor _IMPL((pixel), Pixel pixel )
{
    XColor *color = CMmap_find_pixel(pixel);

    if ( !color ) {
	String          params = (String) pixel;
	Cardinal	num_params = 1;
	XtWarningMsg("CMFreeColor", "badPixel", XtCCMError,
		     "Pixel %d not in managed map",
		     &params, &num_params);
    } else {
	CMmap_free_idx(color-CMmap.color,True);
    }
}

void
CMFreeColors _IMPL((pixels, npixels), Pixel *pixels _AND int npixels )
{
    int i;

    DEBUG((DBG_CM,"CMFreeColors(%d)\n",npixels));
    for ( i = 0; i < npixels; ++i )
      CMFreeColor( pixels[i] );
}

/* 
 * These converters are reinitialized to use the color map manager.
 * 
 */

#define XtNwrongParameters "wrongParameters"
#define XtNconversionError "conversionError"

#define	done(type, value) \
	{							\
	    if (toVal->addr != NULL) {				\
		if (toVal->size < sizeof(type)) {		\
		    toVal->size = sizeof(type);			\
		    return False;				\
		}						\
		*(type*)(toVal->addr) = (value);		\
	    }							\
	    else {						\
		static type static_val;				\
		static_val = (value);				\
		toVal->addr = (XtPointer)&static_val;		\
	    }							\
	    toVal->size = sizeof(type);				\
	    return True;					\
	}

static Boolean 
CvtIntOrPixelToXColor(dpy, args, num_args, fromVal, toVal, closure_ret)
    Display*	dpy;
    XrmValuePtr args;
    Cardinal    *num_args;
    XrmValuePtr	fromVal;
    XrmValuePtr	toVal;
    XtPointer	*closure_ret;
{    
    XColor	c;

    if (*num_args != 0)
      XtErrorMsg("cvtIntOrPixelToXColor",XtNwrongParameters,XtCCMError,
		 "CM Pixel to color conversion takes no arguments",
		 (String *)NULL, (Cardinal *)NULL);
    c.pixel = *(int *)fromVal->addr;

    XQueryColor(CMdisplay, CMcmap, &c);
    done(XColor, c);
}

static Boolean CvtStringToPixel(dpy, args, num_args, fromVal, toVal, closure_ret)
    Display*	dpy;
    XrmValuePtr args;
    Cardinal    *num_args;
    XrmValuePtr	fromVal;
    XrmValuePtr	toVal;
    XtPointer	*closure_ret;
{
    String	    str = (String)fromVal->addr;
    XColor	    screenColor;
    XColor	    exactColor;
    Status	    status;

    if (*num_args != 0)
     XtErrorMsg("cvtStringToPixel", XtNwrongParameters, XtCCMError,
		"CM String to pixel conversion takes no arguments",
		(String *)NULL, (Cardinal *)NULL);

    if (XmuCompareISOLatin1(str, XtDefaultBackground) == 0) {
	*closure_ret = False;
	done(Pixel, WhitePixelOfScreen(CMscreen));
    }
    if (XmuCompareISOLatin1(str, XtDefaultForeground) == 0) {
	*closure_ret = False;
	done(Pixel, BlackPixelOfScreen(CMscreen));
    }

    status = CMAllocNamedColor((char*)str,&screenColor, &exactColor);
    if ( status == 0 ) {
	*closure_ret = (XtPointer)False;
	return False;
    } else {
	*closure_ret = (XtPointer)True;
        done(Pixel, screenColor.pixel);
    }
}

/* ARGSUSED */
static void FreePixel(app, toVal, closure, args, num_args)
    XtAppContext app;
    XrmValuePtr	toVal;
    XtPointer	closure;
    XrmValuePtr	args;
    Cardinal	*num_args;
{
    if (*num_args != 0)
     XtAppErrorMsg(app, XtNwrongParameters,"freePixel",XtCCMError,
	"CM Freeing a pixel takes no arguments",
        (String *)NULL, (Cardinal *)NULL);

    if (closure) {
	CMFreeColors((Pixel*)toVal->addr, 1);
    }
}

static void
InitializeConverters()
{
    XtSetTypeConverter(XtRString, XtRPixel, CvtStringToPixel, NULL, 0,
		       XtCacheByDisplay, FreePixel);
    XtSetTypeConverter(XtRPixel, XtRColor, CvtIntOrPixelToXColor, NULL, 0,
		       XtCacheByDisplay, 0);
}

/* 
 * This stuff has been lifted straight from Jim Frost's xloadimage.  
 * See file "jfcopyright.h".
 */

#include "jfcopyright.h"

static struct visual_class_name {
  int   class; /* numerical value of class */
  char *name;  /* actual name of class */
} VisualClassName[] = {
  TrueColor,   "TrueColor",
  DirectColor, "DirectColor",
  PseudoColor, "PseudoColor",
  StaticColor, "StaticColor",
  GrayScale,   "GrayScale",
  StaticGray,  "StaticGray",
  StaticGray,  "StaticGrey",
  -1,          NULL
};

int 
visualClassFromName _IMPL((name), char *name)
{
    int a;
    char *s1, *s2;
    int class= -1;
    
    if ( !name || *name == '\0' ) return -1;
    for (a= 0; VisualClassName[a].name; a++) {
	for (s1= VisualClassName[a].name, s2= name; *s1 && *s2; s1++, s2++)
	  if ((isupper(*s1) ? tolower(*s1) : *s1) !=
	      (isupper(*s2) ? tolower(*s2) : *s2))
	    break;
	
	if ((*s1 == '\0') || (*s2 == '\0')) {
	    
	    /* check for uniqueness.  we special-case StaticGray 
	     * because we have two spellings but they are unique if 
	     * either is found.
	     */
	    
	    if ((class != -1) && (class != StaticGray)) {
		fprintf(stderr, "%s does not uniquely describe a visual class (ignored)\n", name);
		return(-1);
	    }
	    class= VisualClassName[a].class;
	}
    }
    if ( class == -1 )
      fprintf(stderr, "%s is not a visual class (ignored)\n", name);
    return(class);
}

char *
nameOfVisualClass _IMPL((class), int class)
{
    int a;
    
    for (a= 0; VisualClassName[a].name; a++)
      if (VisualClassName[a].class == class)
	return(VisualClassName[a].name);
    return("[Unknown Visual Class]");
}

/* find the best visual of a particular class with a particular depth
 */

static Visual *
bestVisualOfClassAndDepth _IMPL((disp, screen, class, depth),
				Display *disp _AND Screen *screen _AND 
				int class _AND unsigned int depth)
{
    Visual *best= NULL;
    XVisualInfo template, *info;
    int nvisuals;
    
    template.screen= XScreenNumberOfScreen(screen);
    template.class= class;
    template.depth= depth;
    if (! (info= XGetVisualInfo(disp, VisualScreenMask | VisualClassMask |
				VisualDepthMask, &template, &nvisuals)))
      return(NULL); /* no visuals of this depth */
    
    /* not sure what to do if this gives more than one visual of a particular
     * class and depth, so just return the first one.
     */
    
    best= info->visual;
    XFree((char *)info);
    return(best);
}

/* given a visual class, try to find the best visual of that class at
 * the best depth.  returns a null visual and depth if it couldn't find
 * any visual of that type at any depth
 */

static void
bestVisualOfClass _IMPL((disp, screen, visual_class, rvisual, rdepth),
			Display *disp _AND Screen *screen _AND 
			int visual_class _AND 
			Visual **rvisual _AND unsigned int *rdepth)
{
    Visual       *visual;
    unsigned int  a, b, depth;
    
    /* loop through depths looking for a visual of a good depth which matches
     * our visual class.
     */
    
    visual= (Visual *)NULL;
    depth= 0;
    for (a= 0; a < screen->ndepths; a++) {
	for (b= 0; b < screen->depths[a].nvisuals; b++) {
	    if ((screen->depths[a].visuals[b].class == visual_class) &&
		(!depth || (screen->depths[a].depth > depth) )) {
		depth= screen->depths[a].depth;
		visual= &(screen->depths[a].visuals[b]);
	    }
	}
    }
    *rvisual= visual;
    *rdepth= depth;
}

