/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: GenImage.c,v 0.9.1.4 1991/12/16 20:51:35 leei Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include <sys/types.h>
/*	#include <sys/time.h>		included via "GenericImageP.h"	*/

#include "GenImageP.h"
#include "debug.h"
#include "proto.h"

#include <X11/Quarks.h>

XrmQuark imStringQ = -1;

static XtResource resources[] = {
#define offset(field) XtOffset(GenericImage, generic.field)
    /* {name, class, type, size, offset, default_type, default_addr}, */
    { ImNwidth, ImCWidth, XtRShort, sizeof(short),
          offset(width), XtRImmediate, (XtPointer) 1 },
    { ImNheight, ImCHeight, XtRShort, sizeof(short),
          offset(height), XtRImmediate, (XtPointer) 1 },
    { ImNident, ImCIdent, XtRString, sizeof(char *),
          offset(ident), XtRString, "<unknown>" }
#undef offset
};

static XtResource disp_resources[] = {
#define offset(field) XtOffset(GenericDisplayParams, generic.field)
    /* {name, class, type, size, offset, default_type, default_addr}, */
    { ImNvisible, ImCVisible, XtRBoolean, sizeof(Boolean),
	offset(visible), XtRString, (XtPointer) "True" }
#undef offset
};

static void
  GenericImInitialize __(( GenericImage im, ArgList argv, Cardinal argc ));
static void
  GenericImClassPartInitialize __((ImageClass class));
static void
  GenericImInitializeParams __(( GenericDisplayParams disp ));
static void
  GenericImDestroyParams __(( GenericDisplayParams disp ));
static void
  GenericImRememberParams __(( GenericDisplayParams disp ));
static Boolean
  GenericImSameParams __(( GenericDisplayParams disp ));
static void 
  GenericImDescription __(( GenericImage im, char *buf, int buf_len ));
static void
  GenericImDestroy __(( GenericImage im ));

GenericImageClassRec genericImageClassRec = {
  { /* core fields */
    /* superclass		*/	NULL,
    /* class_name		*/	"Image",
    /* image_size		*/	sizeof(GenericImageRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	GenericImClassPartInitialize,
    /* class_inited		*/	FALSE,
    /* initialize		*/	GenericImInitialize,
    /* resources                */	resources,
    /* num_resources            */	XtNumber(resources),
    /* set_values               */      NULL,
    /* read			*/	NULL,
    /* write			*/	NULL,
    /* image_display_size	*/	sizeof(GenericDisplayParamsRec),
    /* param_resources          */	disp_resources,
    /* num_param_resources      */	XtNumber(disp_resources),
    /* initialize_params	*/	GenericImInitializeParams,
    /* set_params		*/	NULL,
    /* destroy_params		*/	GenericImDestroyParams,
    /* edit_params		*/	NULL,
    /* remember_params		*/	GenericImRememberParams,
    /* same_params		*/	GenericImSameParams,
    /* install			*/	NULL,
    /* uninstall		*/	NULL,
    /* alloc_colors		*/	NULL,
    /* free_colors		*/	NULL,
    /* display			*/	NULL,
    /* description		*/	GenericImDescription,
    /* element_description	*/	NULL,
    /* destroy			*/	GenericImDestroy
  }
};

ImageClass genericImageClass = &genericImageClassRec;


/* 
 * Image and class information access and trivial modification.
 */

ImageClass
ImClass _IMPL((im), GenericImage im )
{
    return im->generic.class;
}

Boolean
ImIsSubclass _IMPL((im, cl), GenericImage im _AND ImageClass cl )
{
    ImageClass scl;

    for ( scl = im->generic.class; scl; scl = ImClassSuperclass(scl) )
      if ( scl == cl ) return True;
    return False;
}

_CONST char *
ImIdent _IMPL((im), GenericImage im )
{
    return im->generic.ident;
}

int
ImWidth _IMPL((im), GenericImage im )
{
    return im->generic.width;
}

int
ImHeight _IMPL((im), GenericImage im )
{
    return im->generic.height;
}

Boolean
ImChanged _IMPL((im), GenericImage im )
{
    return im->generic.changed;
}

_CONST char *
ImComment _IMPL((im), GenericImage im )
{
    return im->generic.comment;
}

_CONST char *
ImSetComment _IMPL((im, comment), 
		   GenericImage im _AND _CONST char *comment )
{
    char *prev_comment = im->generic.comment;
    char *new_comment = malloc( strlen(comment) + 1 );

    free(prev_comment);
    strcpy( new_comment, comment );

    return im->generic.comment = new_comment;
}

_CONST char *
ImAppendComment _IMPL((im, comment, timestamp),
		      GenericImage im _AND _CONST char *comment _AND 
		      Boolean timestamp )
{
    char *prev_comment = im->generic.comment;
    int prev_length = (prev_comment? strlen(prev_comment): 0);
    int new_length = prev_length + (timestamp? 27: 0) + strlen(comment) + 2;
    char *new_comment = malloc( new_length );

    new_comment[0] = '\0';
    if ( prev_comment ) {
	strcpy( new_comment, prev_comment );
	strcat( new_comment, "\n" );
    }
    if ( timestamp ) {
	time_t curr_time;

	time(&curr_time);
	strcat( new_comment, ctime(&curr_time) );
	new_comment[strlen(new_comment)-1] = '\0';
	strcat( new_comment, ": " );
    }
    strcat( new_comment, comment );
    free(prev_comment);

    return im->generic.comment = new_comment;
}

ImageClass
ImClassSuperclass _IMPL((class), ImageClass class )
{
    return class->generic.superclass;
}

_CONST char *
ImClassName _IMPL((class), ImageClass class )
{
    return class->generic.class_name;
}

#define PrognMethod(val,class,type,sel,args)				\
{									\
    type _cl = class;							\
    for ( ; _cl != NULL; _cl = _cl->generic.superclass ) {		\
	if (_cl->sel) { 						\
	    DEBUG((DBG_METH, "%s::%s(%#x)\n", 				\
		  _cl->generic.class_name, _STRING(sel), val));		\
	    (*(_cl->sel))args;						\
	}								\
    }									\
}

#define DaemonMethod(val,class,type,sel,args)				\
{									\
    type _cl = class;							\
    for ( ; _cl != NULL; _cl = _cl->generic.superclass ) {		\
	if (_cl->sel) { 						\
	    DEBUG((DBG_METH, "%s::%s(%#x)\n", 				\
		  _cl->generic.class_name, _STRING(sel), val));		\
	    (*(_cl->sel))args;						\
	    break; 							\
	}								\
    }									\
}

#define DaemonMethodReturn(val,class,type,sel,args)			\
{									\
    type _cl = class;							\
    for ( ; _cl != NULL; _cl = _cl->generic.superclass ) {		\
	if (_cl->sel) {							\
	    DEBUG((DBG_METH, "%s::%s(%#x)\n", 				\
		  _cl->generic.class_name, _STRING(sel), val));		\
	    return (*(_cl->sel))args;					\
	}								\
    }									\
}

#define OrMethod(val,class,type,sel,args)				\
{									\
    type _cl = class;							\
    for ( ; _cl != NULL; _cl = _cl->generic.superclass ) {		\
	if (_cl->sel) DEBUG((DBG_METH, "%s::%s(%#x)\n",			\
			    _cl->generic.class_name, _STRING(sel), val));\
	if (_cl->sel && (*(_cl->sel))args) return True;			\
    }									\
    return False;							\
}

#define AndMethod(val,class,type,sel,args)				\
{									\
    type _cl = class;							\
    for ( ; _cl != NULL; _cl = _cl->generic.superclass ) {		\
	if (_cl->sel) DEBUG((DBG_METH, "%s::%s(%#x)\n",			\
			    _cl->generic.class_name, _STRING(sel), val));\
	if (_cl->sel && !(*(_cl->sel))args) return False;		\
    }									\
    return True;							\
}

static void
  InitializeMethod _IMPL((class, im, argv, argc), 
			 ImageClass class _AND GenericImage im _AND 
			 ArgList argv _AND Cardinal argc )
{
    ImageClass super = class->generic.superclass;

    if ( super ) InitializeMethod(super,im,argv,argc);

    if ( class->generic.initialize ) {
	DEBUG((DBG_METH, "%s::initialize(%#x)\n", class->generic.class_name, im));
	(*class->generic.initialize)(im,argv,argc);
    }
    im->generic.changed = True;
}

void
  ImDestroyImage _IMPL((im), GenericImage im )
{
    if ( !im->generic.params ) {
	PrognMethod( im, im->generic.class, ImageClass, 
		     generic.destroy, (im) );
	free(im);
    }
}

char *
  ImDescription _IMPL((im, buf, buf_len), 
		      GenericImage im _AND char *buf _AND int buf_len )
{
    PrognMethod( im, im->generic.class, ImageClass,
		 generic.description,(im,buf,buf_len) );
    return buf;
}

char *
  ImElementDescription _IMPL((im, i, j, buf, buf_len), 
			     GenericImage im _AND int i _AND int j _AND 
			     char *buf _AND int buf_len )
{
    PrognMethod( im, im->generic.class, ImageClass,
		 generic.element_description,(im,i,j,buf,buf_len) );
    return buf;
}

void
  ImDisplay _IMPL((disp), GenericDisplayParams disp )
{
    Boolean same_params;
    ImageClass _cl = disp->generic.image->generic.class;

    if ( !disp->generic.visible ) return;

    same_params = ImSameParams(disp);
    for ( ; _cl != NULL; _cl = _cl->generic.superclass ) {
	if ( _cl->generic.display ) { 
	    (*_cl->generic.display)(disp,same_params); 
	    break;
	}
    }
    ImRememberParams(disp);
    disp->generic.image->generic.changed = False;
}

void
  ImAllocColors _IMPL((disp), GenericDisplayParams disp )
{
    ImageClass im_cl = (disp->generic.image)->generic.class;

    if ( !disp->generic.visible ) return;

    PrognMethod( disp, im_cl, ImageClass, generic.alloc_colors, (disp) );
}

void
  ImFreeColors _IMPL((disp), GenericDisplayParams disp )
{
    ImageClass im_cl = (disp->generic.image)->generic.class;
    PrognMethod( disp, im_cl, ImageClass, generic.free_colors, (disp) );
}

static void
  DisplayInitializeMethod _IMPL((class, disp),
				ImageClass class _AND 
				GenericDisplayParams disp )
{
    ImageClass super = class->generic.superclass;

    if ( super ) DisplayInitializeMethod(super,disp);

    if ( class->generic.initialize_params ) {
	DEBUG((DBG_METH, "%s::initialize_params(%#x)\n", 
	      class->generic.class_name, disp));
	(*class->generic.initialize_params)(disp);
    }
}

ExternalParams *
  ImExternalParams _IMPL((disp), GenericDisplayParams disp)
{
    return &(disp->xparams);
}

static void
  DestroyParamsMethod _IMPL((class, disp), 
			    ImageClass class _AND GenericDisplayParams disp )
{
    ImageClass super = class->generic.superclass;

    if ( super ) DestroyParamsMethod(super,disp);

    if ( class->generic.destroy_params ) {
	DEBUG((DBG_METH, "%s::destroy_params(%#x)\n", 
	      class->generic.class_name, disp));
	(*class->generic.destroy_params)(disp);
    }
}

static void
AttachParams _IMPL((im, disp), GenericImage im _AND GenericDisplayParams disp)
{
    disp->generic.next = im->generic.params;
    im->generic.params = disp;
}

static void
DetachParams _IMPL((im, disp), GenericImage im _AND GenericDisplayParams disp)
{
    GenericDisplayParams curr = im->generic.params;

    if ( curr == im->generic.params ) {
	im->generic.params = curr->generic.next;
    } else {
	for (; curr && curr->generic.next != disp; curr = curr->generic.next);
	if ( !curr ) {
	    fprintf(stderr, "DisplayParams %#x already detached from %#x\n",
		    disp, im);
	    return;
	} else {
	    curr->generic.next = disp->generic.next;
	}
    }
}

void
  ImDestroyParams _IMPL((disp, destroy_image),
			GenericDisplayParams disp _AND Boolean destroy_image)
{
    ImageClass im_cl = (disp->generic.image)->generic.class;
    DestroyParamsMethod(im_cl, disp);

    /* If reference count reduces to zero, destroy. */
    DetachParams(disp->generic.image, disp);
    if ( destroy_image ) ImDestroyImage(disp->generic.image);

    free(disp->generic.prev);
    free(disp);
}

void
  ImRememberParams _IMPL((disp), GenericDisplayParams disp )
{
    ImageClass cl = disp->generic.image->generic.class;
    PrognMethod(disp, cl, ImageClass, generic.remember_params, (disp));
}

Boolean
  ImSameParams _IMPL((disp), GenericDisplayParams disp )
{
    ImageClass cl = disp->generic.image->generic.class;
    AndMethod(disp, cl, ImageClass, generic.same_params, (disp));
}

void
  ImInstall _IMPL((disp), GenericDisplayParams disp )
{
    ImageClass cl = disp->generic.image->generic.class;
    PrognMethod(disp, cl, ImageClass, generic.install, (disp));
}

void
  ImUninstall _IMPL((disp), GenericDisplayParams disp )
{
    ImageClass cl = disp->generic.image->generic.class;
    PrognMethod(disp, cl, ImageClass, generic.uninstall, (disp));
}

/* This is lifted nearly wholesale from the MIT Xt intrinsics, so:
Copyright 1987, 1988 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

#include <X11/IntrinsicP.h>

#if DEBUGGING
static void
describe_resources _IMPL((msg, class, indir),
			 char *msg _AND ImageClass class _AND Boolean indir)
{
    int i;
    XrmResourceList resL = (XrmResourceList) class->generic.resources;
    XrmResourceList *resLL = (XrmResourceList *)class->generic.resources;

    fprintf(stderr, "Class %s resources %#x %d: %s\n",
	    class->generic.class_name, class->generic.resources,
	    class->generic.num_resources, msg);
    for ( i = 0; i < class->generic.num_resources; ++i ) {
	XrmResourceList res = indir? resLL[i]: &resL[i];
	fprintf(stderr, "Res[%d] %d %s %d %s offset %d\n",
		i, res->xrm_name, XrmQuarkToString(res->xrm_name), 
		res->xrm_class, XrmQuarkToString(res->xrm_class),
		-res->xrm_offset-1);
    }
}
#endif

static void 
  _ImResourceDependencies _IMPL((class), ImageClass class)
{
    ImageClass super;
    XrmResourceList *super_resources = 0;
    int super_num_resources = 0;
    int super_size = 0;
    int i;

    super = class->generic.superclass;

    if ( super ) {
	super_resources = (XrmResourceList *) super->generic.resources;
	super_num_resources = super->generic.num_resources;
	super_size = super->generic.image_size;
    }
    _XtDependencies(&(class->generic.resources),
		    &(class->generic.num_resources),
		    super_resources, super_num_resources, super_size);

    if ( super ) {
	super_resources = (XrmResourceList *) super->generic.param_resources;
	super_num_resources = super->generic.num_param_resources;
	super_size = super->generic.image_display_size;
    }
    _XtDependencies(&(class->generic.param_resources),
		    &(class->generic.num_param_resources),
		    super_resources, super_num_resources, super_size);
}

#include <X11/Xresource.h>

static void
  ClassPartInitializeMethod _IMPL((ancestor, class),
				  ImageClass ancestor _AND ImageClass class )
{
    ImageClass super = ancestor->generic.superclass;
    int i;

    if ( super ) ClassPartInitializeMethod(super,class);

    /* Actually call the part initialization. */
    if ( class->generic.class_part_initialize ) {
	DEBUG((DBG_METH, "%s::class_part_initialize()\n", 
	      class->generic.class_name));
	(*class->generic.class_part_initialize)(class);
    }
}

#ifdef X11R5
#  define XrmCompileResourceList _XtCompileResourceList
#endif

static void
  ImClassInitialize _IMPL((class), ImageClass class )
{
    if ( imStringQ < 0 ) {
#if !defined(XlibSpecificationRelease) || XlibSpecificationRelease < 5
	imStringQ = XrmStringToQuark("String");
#else
	imStringQ = XrmPermStringToQuark("String");
#endif
    }

    /* Ensure that the superclass is initialized before anything else. */
    if ( class->generic.superclass &&
	!class->generic.superclass->generic.class_inited )
      ImClassInitialize(class->generic.superclass);

    /* Compile resource list. */
    if ( class->generic.resources ) {
	XrmCompileResourceList(class->generic.resources,
			       class->generic.num_resources);
    }

    /* Compile parameter resource list. */
    if ( class->generic.param_resources ) {
	XrmCompileResourceList(class->generic.param_resources,
			       class->generic.num_param_resources);
    }

    _ImResourceDependencies(class);

    /* Now initialize the class: part-by-part then as a whole. */
    ClassPartInitializeMethod(class,class);
    DaemonMethod( class, class, ImageClass,
		  generic.class_initialize, (class) );

    class->generic.class_inited = TRUE;
}

static Boolean
  SetValuesMethod _IMPL((class, old, request, new),
			ImageClass class _AND 
			GenericImage old _AND 
			GenericImage request _AND
			GenericImage new )
{
    ImageClass super = class->generic.superclass;
    Boolean result = FALSE;

    if ( super ) 
      result = SetValuesMethod( super, old, request, new );

    if ( class->generic.set_values ) {
	DEBUG((DBG_METH, "%s::set_values(%#x)\n",class->generic.class_name,new));
	result |= (*class->generic.set_values)(old,request,new);
    }

    return result;
}

static XtResourceList
directResources _IMPL((res, num_res), 
		      XtResourceList res _AND Cardinal num_res )
{
    XrmResourceList *xrmres = (XrmResourceList *) res;
    XrmResourceList dir_res = 
      (XrmResourceList) XtMalloc(num_res * sizeof(XrmResource));
    int i;

    for ( i = 0; i < num_res; ++i ) dir_res[i] = *xrmres[i];

    return (XtResourceList) dir_res;
}

#define SETVALUES

#ifdef SETVALUES
static void SetValues(base, res, num_resources, args, num_args)
  char*			base;		/* Base address to write values to   */
  XrmResourceList*	res;		/* The current resource values.      */
  register Cardinal	num_resources;	/* number of items in resources      */
  ArgList 		args;		/* The resource values to set        */
  Cardinal		num_args;	/* number of items in arg list       */
{
    register ArgList		arg;
    register int 	        i;
    register XrmName		argName;
    register XrmResourceList*   xrmres;

    /* Resource lists are assumed to be in compiled form already via the
       initial XtGetResources, XtGetSubresources calls */

    for (arg = args ; num_args != 0; num_args--, arg++) {
	argName = XrmStringToName(arg->name);
	for (xrmres = res, i = 0; i < num_resources; i++, xrmres++) {
	    if (argName == (*xrmres)->xrm_name) {
		_XtCopyFromArg(arg->value,
		    base - (*xrmres)->xrm_offset - 1,
		    (*xrmres)->xrm_size);
		break;
	    }
	}
    }
} /* SetValues */
#endif

void
  ImSetValues _IMPL((im, argv, argc),
		    GenericImage im _AND ArgList argv _AND Cardinal argc )
{
    GenericImage old, req;
    ImageClass class = im->generic.class;
    Cardinal image_size = class->generic.image_size;
    XtResourceList res;
    Cardinal num_res = class->generic.num_resources;

    if ( num_res ) {
	old = (GenericImage) alloca( image_size  );
	req = (GenericImage) alloca( image_size );
	
	memcpy( old, im, image_size );
#ifdef SETVALUES
	SetValues((XtPointer)im, class->generic.resources, num_res, 
		  argv, argc);
#else
	res = directResources(class->generic.resources, num_res);
	XtSetSubvalues((XtPointer) im, res, num_res, argv, argc);
	XtFree((char *)res);
#endif
	memcpy( req, im, image_size );

	im->generic.changed |= SetValuesMethod( class, old, req, im );
    }
}

void
  ImPutProperties _IMPL((im, propv, propc),
			GenericImage im _AND 
			ImPropList propv _AND Cardinal propc )
{
    char * base = (char *) im;
    GenericImage old, req;
    ImageClass class = im->generic.class;
    Cardinal image_size = class->generic.image_size;
    ImageProperty prop = im->generic.props;
    ImageProperty prev;

    int i;
    XrmName argName, argType; XrmResourceList *xrmres;
    Boolean changed = False;

    old = (GenericImage) alloca( image_size  );
    req = (GenericImage) alloca( image_size );

    memcpy( old, im, image_size );
    for ( ; propc; ++propv, --propc ) {
	/* 
	 * Resource lists are assumed to be in compiled form already via the
	 * initial XtGetResources, XtGetSubresources calls
	 */
	argName = XrmStringToQuark(propv->name);
	argType = XrmStringToQuark(propv->type);
	for (xrmres = (XrmResourceList *) class->generic.resources, i = 0; 
	     i < class->generic.num_resources; i++, xrmres++) {
	    if (argName == (*xrmres)->xrm_name) {
		XrmValue from, to;
		to.size = (*xrmres)->xrm_size;
		to.addr = base - (*xrmres)->xrm_offset - 1;
		from.size = propv->value.size; 
		if ((argType == imStringQ) || (to.size > sizeof(XtArgVal)))
		  from.addr = (caddr_t) propv->value.addr;
		else
		  _XtCopyFromArg( propv->value.addr, &from.addr, from.size );
		ImConvert(argType, &from, (*xrmres)->xrm_type, &to);
		changed = True;
		goto cont;
	    }
	}

	/* If resource not found, then add a property. */
	for ( prev = NULL; prop; prev = prop, prop = prop->next ) 
	  if ( prop->name == argName ) break;

	if ( !prop )
	  prop = (ImageProperty) calloc( sizeof(*prop), 1 );
	if ( prev ) 
	  prev->next = prop;
	else 
	  im->generic.props = prop;

	/* Initialize the record. */
	prop->name = argName;
	prop->type = argType;
	prop->data.size = propv->value.size;
	_XtCopyFromArg(propv->value.addr, &prop->data.addr, prop->data.size);
      cont:;
    }
    memcpy( req, im, image_size );

    if ( changed )
      im->generic.changed |= SetValuesMethod( class, old, req, im );
}

Boolean
  ImGetProperties _IMPL((im, propv, propc),
			GenericImage im _AND 
			ImPropList propv _AND Cardinal propc )
{
    ImageClass class = im->generic.class;
    ImageProperty prop = im->generic.props;
    ImageProperty prev;
    Boolean success = True;

    int i;
    XrmName argName, argType; XrmResourceList *xrmres;

    for ( ; propc; ++propv, --propc ) {
	/* 
	 * Resource lists are assumed to be in compiled form already via the
	 * initial XtGetResources, XtGetSubresources calls
	 */
	argName = XrmStringToQuark(propv->name);
	argType = XrmStringToQuark(propv->type);
	for (xrmres = (XrmResourceList *) class->generic.resources, i = 0; 
	     i < class->generic.num_resources; i++, xrmres++) {
	    if (argName == (*xrmres)->xrm_name) {
		XrmValue from, to;
		from.size = (*xrmres)->xrm_size;
		from.addr = ((char *) im) - (*xrmres)->xrm_offset - 1;
		to.addr = propv->value.addr;
		ImConvert((*xrmres)->xrm_type, &from, argType, &to );
		goto cont;
	    }
	}
	
	/* If resource not found, then try properties. */
	for ( prev = NULL; prop; prev = prop, prop = prop->next ) 
	  if ( prop->name == argName ) break;

	if ( prop ) {
	    XrmValue from, to;
	    from.size = prop->data.size; 
	    if ((prop->type == imStringQ) || (from.size > sizeof(XtArgVal)))
	      from.addr = (caddr_t) prop->data.addr;
	    else
	      from.addr = (caddr_t) &prop->data.addr;
	    to.addr = propv->value.addr;
	    to.size = propv->value.size;
	    ImConvert(prop->type, &from, argType, &to);
	}
	else
	  success = False;
      cont:;
    }
    return success;
}

GenericImage
  ImCreateImage _IMPL((id, class, w, h, argv, argc), 
		      char *id _AND ImageClass class _AND int w _AND int h _AND
		      ArgList argv _AND Cardinal argc )
{
    GenericImage im;
    int i;

    if ( !class->generic.class_inited )
      ImClassInitialize(class);

    im = (GenericImage) calloc(1,class->generic.image_size);
    DEBUG((DBG_IM, "CreateImage(%s, %#x, %dx%d) = %#x (%d)\n", id, class, w, h,
	   im, class->generic.image_size));
    if ( !im ) return im;

    im->generic.class   = class;
    im->generic.width   = w;
    im->generic.height  = h;

    if ( class->generic.num_resources ) {
	Cardinal num_res = class->generic.num_resources;
#ifdef SETVALUES
	SetValues((XtPointer)im, class->generic.resources, num_res, 
		  argv, argc);
#else
	XtResourceList res = 
	  directResources(class->generic.resources, num_res);
	XtSetSubvalues((XtPointer) im, res, num_res, argv, argc);
	XtFree((char *)res);
#endif
    }

    /* Make a local copy of the identifier. */
    im->generic.ident   = malloc(strlen(id)+1);
    strcpy(im->generic.ident,id);

    InitializeMethod(class,im,argv,argc);

    return im;
}

GenericDisplayParams
  ImCreateParams _IMPL((im, xparams),
		       GenericImage im _AND ExternalParams *xparams)
{
    ImageClass class = im->generic.class;

    GenericDisplayParams disp = 
      (GenericDisplayParams) calloc(1,class->generic.image_display_size);
    DEBUG((DBG_IM, "CreateParams(%#x %s) = %#x (%d)\n", im, ImIdent(im), disp,
	   class->generic.image_display_size));
    if ( !disp ) return disp;

    if ( !class->generic.class_inited )
      ImClassInitialize(class);

    disp->generic.image = im;
    disp->generic.prev =
      (GenericDisplayParams) calloc(1,class->generic.image_display_size);
    if ( xparams ) disp->xparams = *xparams;

    /* Increment reference count. */
    AttachParams(im, disp);

    /* 
     * This is not a full resource facility like the X Toolkit, all 
     * relevant arguments must be passed in here in the same format as 
     * for an ImSetParams(...) list.
     */
    if ( class->generic.num_param_resources ) {
	String class_name = (String)
	  alloca( strlen(class->generic.class_name)+8 );
	Cardinal num_res = class->generic.num_param_resources;
	XtResourceList res = 
	  directResources(class->generic.param_resources, num_res);
	strcpy(class_name, class->generic.class_name);
	strcat(class_name, "Display");
	DEBUG((DBG_IM, "GetSubresources(%s, %#x, %s, %s, ...)\n",
	       XtName(disp->xparams.widget), disp,
	       im->generic.ident, class_name));
	XtGetSubresources(disp->xparams.widget,
			  (XtPointer) disp, im->generic.ident, class_name,
			  res, num_res,
			  NULL, 0 );
	XtFree((char *)res);
    }

    DisplayInitializeMethod(class, disp);
    return disp;
}

static void
  GenericImClassPartInitialize _IMPL((class), ImageClass class)
{
}

static void
  GenericImInitialize _IMPL((im, argv, argc),
			    GenericImage im _AND 
			    ArgList argv _AND Cardinal argc )
{
    im->generic.props = NULL;
    im->generic.params = NULL;
}

static void
  GenericImDestroy _IMPL((im), GenericImage im )
{
    ImageProperty prop, next;

    ImRemoveImage(im);

    if ( im->generic.ident ) {
	free(im->generic.ident);
	im->generic.ident = 0;
    }
    if ( im->generic.comment ) {
	free(im->generic.comment);
	im->generic.comment = 0;
    }
    if ( im->generic.file_name ) {
	free(im->generic.file_name);
	im->generic.file_name = 0;
    }

    for ( prop = im->generic.props; prop; prop = next ) {
	next = prop->next;
	free( prop );
    }
    im->generic.props = 0;
}

static void
  GenericImDescription _IMPL((im, buf, buf_len),
			     GenericImage im _AND char *buf _AND int buf_len )
{
    sprintf( buf, "%s %s (%dx%d)",
	    ImClassName(ImClass(im)),
	    ImIdent(im),
	    ImWidth(im), ImHeight(im));
}

static void
  GenericImInitializeParams _IMPL((disp), GenericDisplayParams disp)
{
    /* No previous pixmap. */
    disp->generic.pm = 0;
}

static Boolean
  SetParamsMethod _IMPL((class, old, request, new),
			ImageClass class _AND
			GenericDisplayParams old _AND 
			GenericDisplayParams request _AND
			GenericDisplayParams new )
{
    ImageClass super = class->generic.superclass;
    Boolean result = FALSE;

    if ( super ) 
      result = SetParamsMethod( super, old, request, new );

    if ( class->generic.set_params ) {
	DEBUG((DBG_METH, "%s::set_params(%#x)\n",class->generic.class_name,new));
	result |= (*class->generic.set_params)(old,request,new);
    }

    return result;
}

Boolean
  ImSetParams _IMPL((disp, argv, argc),
		    GenericDisplayParams disp _AND
		    ArgList argv _AND Cardinal argc )
{
    GenericDisplayParams old, req;
    GenericImage im = disp->generic.image;
    ImageClass class = im->generic.class;
    Cardinal display_size = class->generic.image_display_size;
    Boolean changed = False;
    XtResourceList res;
    Cardinal num_res = class->generic.num_param_resources;

    if ( num_res ) {
	old = (GenericDisplayParams) alloca( display_size );
	req = (GenericDisplayParams) alloca( display_size );

	memcpy( old, disp, display_size );
#ifdef SETVALUES
	SetValues((XtPointer)disp, class->generic.param_resources, num_res, 
		  argv, argc);
#else
	res = directResources(class->generic.param_resources, num_res);
	XtSetSubvalues((XtPointer)disp, res, num_res, argv, argc);
	XtFree((char *)res);
#endif
	memcpy( req, disp, display_size );

	if ( SetParamsMethod( class, old, req, disp ) )
	  ImRedraw(disp->xparams.widget, False);
    }
}

Boolean
  ImEditParams _IMPL((disp), GenericDisplayParams disp)
{
    GenericDisplayParams old, req;
    ImageClass cl = disp->generic.image->generic.class;
    Cardinal display_size = cl->generic.image_display_size;

    old = (GenericDisplayParams) alloca( display_size );
    req = (GenericDisplayParams) alloca( display_size );

    memcpy( old, disp, display_size );
    DaemonMethod( disp, cl, ImageClass, generic.edit_params,(disp) );
    memcpy( req, disp, display_size );

    if ( SetParamsMethod( cl, old, req, disp ) )
      ImRedraw(disp->xparams.widget, False);
}

static void
  GenericImDestroyParams _IMPL((disp), GenericDisplayParams disp)
{
    /* No previous pixmap. */
    ImFreeColors(disp);
    if ( disp->generic.pm ) {
	XFreePixmap(disp->xparams.display, disp->generic.pm);
	disp->generic.pm = 0;
    }
}

static void
  GenericImRememberParams _IMPL((disp), GenericDisplayParams disp )
{
    GenericDisplayParams prev = disp->generic.prev;

    *prev = *disp;
}

static Boolean
  GenericImSameParams _IMPL((disp), GenericDisplayParams disp )
{
    GenericDisplayParams prev = disp->generic.prev;
    ExternalParams *cxparams = ImExternalParams(disp);
    ExternalParams *pxparams = ImExternalParams(prev);

    return
      !ImChanged(disp->generic.image) &&
      (disp->generic.visible == prev->generic.visible) &&
      (pxparams->display == cxparams->display) &&
      (pxparams->screen  == cxparams->screen) &&
      (pxparams->visual  == cxparams->visual) &&
      (pxparams->win     == cxparams->win) &&
      XEqualRegion(pxparams->region,cxparams->region) &&
      (pxparams->mag_x   == cxparams->mag_x) &&
      (pxparams->mag_y   == cxparams->mag_y) &&
      (pxparams->range.x  == cxparams->range.x) &&
      (pxparams->range.y  == cxparams->range.y) &&
      (pxparams->range.width  == cxparams->range.width) &&
      (pxparams->range.height  == cxparams->range.height);
}

/*
 * When an image has changed, this function may be called to update 
 * all relevant displays.
 */

void
  ImUpdateImage _IMPL((im, range),
		      GenericImage im _AND XRectangle *range)
{
    GenericDisplayParams disp;

    ++im->generic.changed;
    for ( disp = im->generic.params; disp; disp = disp->generic.next ) {
	if ( range ) {
	    disp->xparams.range = *range;
	} else {
	    disp->xparams.range.x = 0;
	    disp->xparams.range.y = 0;
	    disp->xparams.range.width = im->generic.width;
	    disp->xparams.range.height = im->generic.height;
	}
	ImAllocColors(disp);
	ImDisplay(disp);
    }
}

Boolean
ImConvert _IMPL((type1, val1, type2, val2),
		XrmQuark type1 _AND XrmValuePtr val1 _AND 
		XrmQuark type2 _AND XrmValuePtr val2 )
{
    extern Widget toplevel;

    if ( type1 == type2 ) {
	if ((type1 == imStringQ) || (val1->size > sizeof(XtArgVal)))
	  memcpy( val2->addr, &val1->addr, sizeof(XtArgVal) );
	else
	  memcpy( val2->addr, val1->addr, val1->size );
	return True;
    } else {
	return _XtConvert(toplevel, type1, val1, type2, val2, NULL);
    }
}

ImFileAssoc imFileList = 0;
int imFileListLength = 0;
static int imFileListSize = 0;

void
ImAddFile _IMPL((image), GenericImage image )
{
    char *filename = image->generic.file_name;
    ImFileAssoc assoc;

    if ( !filename ) return;
    DEBUG((DBG_IM, "AddFile(%#x %s) %s\n", image, ImIdent(image), filename));
    assoc = ImFindFile(filename);

    if ( assoc ) {
	if ( assoc->image->generic.width == image->generic.width &&
	     assoc->image->generic.height == image->generic.height) {
	    /* Protect against trashing current assoc. */
	    GenericImage old_im = assoc->image;
	    GenericDisplayParams disp = old_im->generic.params;

	    assoc->filename = filename;
	    assoc->image = image;
	    
	    /* Existing DisplayParams inherit new image. */
	    image->generic.params = old_im->generic.params;
	    for ( ; disp; disp = disp->generic.next  ) {
		disp->generic.image = image;
	    }
	    old_im->generic.params = 0;
	    ImDestroyImage(old_im);
	}
    } else {
	++imFileListLength;
	if ( imFileListLength > imFileListSize ) {
	    imFileListSize += 10;
	    imFileList = (ImFileAssoc) 
	      XtRealloc((char*) imFileList, 
			imFileListSize*sizeof(ImFileAssocRec));
	}
	assoc = &imFileList[imFileListLength-1];
	assoc->filename = filename;
	assoc->image = image;
	DEBUG((DBG_IM, "AddFile(%#x %s) at %d\n", image, ImIdent(image), 
	       imFileListLength-1));

    }
}

void
ImRemoveImage _IMPL((image), GenericImage image )
{
    char *filename = image->generic.file_name;
    ImFileAssoc assoc;
    int i;

    if ( !filename ) return;
    DEBUG((DBG_IM, "RemoveFile(%#x %s) %s\n", image, ImIdent(image), filename));

    assoc = ImFindFile(filename);
    if ( !assoc ) return;
    
    for ( i = assoc-imFileList; i < imFileListLength-1; ++i, ++assoc )
      assoc[0] = assoc[1];

    --imFileListLength;
    imFileList[imFileListLength].filename = 0;
    imFileList[imFileListLength].image = 0;
}

ImFileAssoc
ImFindFile _IMPL((filename), char *filename )
{
    ImFileAssoc assoc;
    int i;

    for ( i = 0, assoc = imFileList ; i < imFileListLength; ++i, ++assoc ) 
      if ( assoc->filename && !strcmp(filename, assoc->filename) ) {
	  DEBUG((DBG_IM, "FindFile(%s) at %d %#x\n", filename, i, assoc->image));
	  return assoc;
      }

    DEBUG((DBG_IM, "FindFile(%s) not found!\n", filename));
    return NULL;
}
