/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: gamma.c,v 0.9.1.3 1991/12/16 21:02:52 leei Exp $";

#include <stdio.h>
#include <math.h>

#include "gamma.h"
#include "proto.h"
#include "utils.h"

#ifndef _POSIX_SOURCE                   /* this is !POSIX */
#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif  /* !M_PI */
#ifdef hp9000
#define fmod(x,y)	drem((x),(y))
#endif	/* hp9000 */
#endif  /* !_POSIX_SOURCE */

/* 
 * COLOR MODELS.
 *   Translations back and forth between a XColor values and a number 
 *   of different color models.
 *   
 *   IHS - Intensity/Hue/Saturation.  A standard, but floating-point 
 *         intensive color model, which captures the intensity as 
 *         the norm of the (r,g,b) vector.
 *
 *   HexIHS - A linear approximation to the IHS model.  Much faster 
 *         to compute since it doesn't involve any transcendental 
 *         arithmetic.
 *              
 *   YIQ - Yowch.  I don't find it particularly useful.
 *   
 *   RGB - Standard translated into float 0.0 -> 1.0 range.
 *  
 */

#define CLIP(x,min,max) ((x)<(min)? (min): (x)>(max)? (max): (x))

#define XColor_float(x)	((x)/65535.0)

static unsigned short
float_XColor _IMPL((x), float x)
{
    return (x<0)? 0: (x>1)? 65535: (unsigned short)(x*65535.0);
}

static float
max3 _IMPL((a, b, c), float a _AND float b _AND float c )
{
    if ( a > b ) b = a;
    return b > c? b: c;
}

static float
min3 _IMPL((a, b, c), float a _AND float b _AND float c )
{
    if ( a < b ) b = a;
    return b < c? b: c;
}

/*
 * This IHS is the hexcone-IHS of Alvy Ray Smith, Siggraph (1978).
 */

void
RGBtoHexIHS _IMPL((in, out), XColor *in _AND Color *out )
{
    float r = XColor_float(in->red);
    float g = XColor_float(in->green);
    float b = XColor_float(in->blue);
    
    if ( r == 0.0 && g == 0.0 && b == 0.0 ) {
	out->ihs.intens = out->ihs.hue = out->ihs.sat = 0.0;
    } else {
	float v = max3(r,g,b);
	float x = min3(r,g,b);
	float sat = (v-x)/v;

	out->ihs.intens = v;
	out->ihs.sat = sat;
	if ( sat == 0.0 ) {
	    out->ihs.hue = 0.0;
	} else {
	    float lr = (v-r)/(v-x);
	    float lg = (v-g)/(v-x);
	    float lb = (v-b)/(v-x);
	    float hh = ((r==v)? ((g==x)? 5+lb: 1-lg):
			(g==v)? ((b==x)? 1+lr: 3-lb):
			((r==x)? 3+lg: 5-lr));
	    out->ihs.hue = hh/6.0;
	}
    }
}

void
HexIHStoRGB _IMPL((in, out), Color *in _AND XColor *out )
{
    float v = in->ihs.intens;
    float h = in->ihs.hue;
    float s = in->ihs.sat;
    int xint = (int) (6*h);
    float f = fmod(6*h,1.0);
    float m = v*(1-s);
    float n = v*(1-s*(xint&1? f: 1-f));

    switch (xint) {
      case 0:
      case 6:
	out->red   = float_XColor(v);
	out->green = float_XColor(n);
	out->blue  = float_XColor(m);
	break;
      case 1:
	out->red   = float_XColor(n);
	out->green = float_XColor(v);
	out->blue  = float_XColor(m);
	break;
      case 2:
	out->red   = float_XColor(m);
	out->green = float_XColor(v);
	out->blue  = float_XColor(n);
	break;
      case 3:
	out->red   = float_XColor(m);
	out->green = float_XColor(n);
	out->blue  = float_XColor(v);
	break;
      case 4:
	out->red   = float_XColor(n);
	out->green = float_XColor(m);
	out->blue  = float_XColor(v);
	break;
      case 5:
	out->red   = float_XColor(v);
	out->green = float_XColor(m);
	out->blue  = float_XColor(n);
	break;
    }
}

static float
HexIHSintens _IMPL((in), Color *in )
{
    return in->ihs.intens;
}

static void
SetHexIHSintens _IMPL((in, intens), Color *in _AND float intens )
{
    in->ihs.intens = intens;
}

static float ihs_c1 = 0.0;
static float ihs_c2 = 0.0;
static float ihs_c3 = 0.0;
static float ihs_int = 0.0;

void 
RGBtoIHS _IMPL((in, out), XColor *in _AND Color *out )
{
    float r = XColor_float(in->red);
    float g = XColor_float(in->green);
    float b = XColor_float(in->blue);

    float x, y, z, intens;

    if ( ihs_c1 == 0.0 ) {
	ihs_c1 = sqrt(1.0/6.0);
	ihs_c2 = sqrt(1.0/2.0);
	ihs_c3 = sqrt(1.0/3.0);
	ihs_int = sqrt(3.0);
    }
    
    x = ihs_c1 * (2*r-b-g);
    y = ihs_c2 * (g-b);
    z = ihs_c3 * (r+g+b);

    out->ihs.intens = intens = sqrt(x*x+y*y+z*z);
    if ( x == 0.0 && y == 0.0 ) {
	out->ihs.hue = out->ihs.sat = 0.0;
    } else {
	float hh = (atan2(y,x) + M_PI)/(2*M_PI);
	float f  = z/intens;

	out->ihs.hue = fmod(hh+0.5,1.0);
	out->ihs.sat = atan2( sqrt(1-f*f), f );
    }
}

void
IHStoRGB _IMPL((in, out), Color *in _AND XColor *out )
{
    float v = in->ihs.intens;
    float h = in->ihs.hue;
    float s = in->ihs.sat;
    
    float hh = 2*M_PI*fmod(h+0.5,1.0) - M_PI;
    float s3 = sin(s);
    float x, y, z;

    if ( ihs_c1 == 0.0 ) {
	ihs_c1 = sqrt(1.0/6.0);
	ihs_c2 = sqrt(1.0/2.0);
	ihs_c3 = sqrt(1.0/3.0);
	ihs_int = sqrt(3.0);
    }
    
    x = ihs_c1 * s3 * cos(hh) * v;
    y = ihs_c2 * s3 * sin(hh) * v;
    z = ihs_c3 * cos(s) * v;

    out->red   = float_XColor( x+x+z );
    out->green = float_XColor( y+z-x );
    out->blue  = float_XColor( z-x-y );
}

static float
IHSintens _IMPL((in), Color *in )
{
    return in->ihs.intens/ihs_int;
}

static void
SetIHSintens _IMPL((in, intens), Color *in _AND float intens )
{
    in->ihs.intens = intens*ihs_int;
}

void 
RGBtoYIQ _IMPL((in, out), XColor *in _AND Color *out )
{
    float r = XColor_float(in->red);
    float g = XColor_float(in->green);
    float b = XColor_float(in->blue);

    out->yiq.y = 0.30*r + 0.59*g + 0.11*b;
    out->yiq.i = 0.60*r - 0.28*g - 0.32*b;
    out->yiq.q = 0.21*r - 0.52*g + 0.31*b;
}

void
YIQtoRGB _IMPL((in, out), Color *in _AND XColor *out )
{
    float y = in->yiq.y;
    float i = in->yiq.i;
    float q = in->yiq.q;

    out->red   = float_XColor(y + 0.948262244*i + 0.62401264*q);
    out->green = float_XColor(y - 0.276066350*i - 0.63981043*q);
    out->blue  = float_XColor(y - 1.105450240*i + 1.72985782*q);
}

static float
YIQintens _IMPL((in), Color *in )
{
    return in->yiq.y;
}

static void
SetYIQintens _IMPL((in, intens), Color *in _AND float intens )
{
    in->yiq.y = intens;
}

void 
RGBtoFRGB _IMPL((in, out), XColor *in _AND Color *out )
{
    out->rgb.r = XColor_float(in->red);
    out->rgb.g = XColor_float(in->green);
    out->rgb.b = XColor_float(in->blue);
}

void
FRGBtoRGB _IMPL((in, out), Color *in _AND XColor *out )
{
    out->red   = float_XColor(in->rgb.r);
    out->green = float_XColor(in->rgb.g);
    out->blue  = float_XColor(in->rgb.b);
}

static float
FRGBintens _IMPL((in), Color *in )
{
    return (in->rgb.r+in->rgb.g+in->rgb.b)/3;
}

static void
SetFRGBintens _IMPL((in, intens), Color *in _AND float intens )
{
    float curr = (in->rgb.r+in->rgb.g+in->rgb.b+0.3)/3;

    if ( in->rgb.r == 0 && in->rgb.g == 0 && in->rgb.b == 0 ) {
	in->rgb.r = in->rgb.g = in->rgb.b = intens;
    } else {
	in->rgb.r = MIN(1.0, (in->rgb.r+.1) * intens/curr);
	in->rgb.g = MIN(1.0, (in->rgb.g+.1) * intens/curr);
	in->rgb.b = MIN(1.0, (in->rgb.b+.1) * intens/curr);
    }
}

static GammaParams defaultGammaParams = {
    YIQ, False,
    { {0.0, 0.0}, {0.25, 0.4}, {0.5, 0.7}, {0.75, 0.9}, {1.0, 1.0} }
};

static GammaParams enhanceGammaParams = {
    YIQ, False,
    { {0.0, 0.0}, {0.25, 0.15}, {0.5, 0.5}, {0.75, 0.85}, {1.0, 1.0} }
};

static GammaParams flatGammaParams = {
    YIQ, False,
    { {0.0, 0.0}, {0.25, 0.25}, {0.5, 0.5}, {0.75, 0.75}, {1.0, 1.0} }
};

static GammaParams invGammaParams = {
    YIQ, False,
    { {0.0, 1.0}, {0.25, 0.75}, {0.5, 0.5}, {0.75, 0.25}, {1.0, 0.0} }
};

static GammaParams dullGammaParams = {
    YIQ, False,
    { {0.0, 0.0}, {0.25, 0.1}, {0.5, 0.3}, {0.75, 0.6}, {1.0, 1.0} }
};

void
defaultGamma _IMPL((params), GammaParams *params )
{
    *params = defaultGammaParams;
}

void
chooseGamma _IMPL((params, which),
		  GammaParams *params _AND WhichGamma which )
{
    switch (which) {
      case gamDefault: *params = defaultGammaParams; break;
      case gamFlat:    *params = flatGammaParams;    break;
      case gamEnhance: *params = enhanceGammaParams; break;
      case gamInverse: *params = invGammaParams;     break;
      case gamDull:    *params = dullGammaParams;    break;
    }
}

void
chooseGammaModel _IMPL((params, which),
		       GammaParams *params _AND ColorModel which )
{
    params->model = which;
}

void 
setGammaExp _IMPL((params, expon),
		  GammaParams *params _AND float expon )
{
    int i;
    float div = exp(expon);

    for ( i = 0; i < 5; ++i ) {
	float x = params->anchors[i].in;
	params->anchors[i].out = x*exp(expon*x)/div;
    }
}

void
dimGamma _IMPL((params, dimmer), 
	       GammaParams *params _AND float dimmer )
{
    int i;

    if ( dimmer < 0 || dimmer > 1.0 ) return;

    for ( i = 0; i < 5; ++i ) {
	float x = params->anchors[i].out;
	float y = (x-dimmer)/(1-dimmer);

	params->anchors[i].out = CLIP(y,0.0,1.0);
    }
}

void
brightenGamma _IMPL((params, brighter),
		    GammaParams *params _AND float brighter )
{
    int i;

    if ( brighter < 0 || brighter > 1.0 ) return;

    for ( i = 0; i < 5; ++i ) {
	float x = params->anchors[i].out;
	float y = x + brighter*(1.0-x);
	params->anchors[i].out = CLIP(y,0.0,1.0);
    }
}

void
dullGamma _IMPL((params, duller),
		GammaParams *params _AND float duller )
{
    float x, y;

    if ( duller < 0 || duller > 1.0 ) return;

    x = params->anchors[1].out;
    y = x + duller;
    params->anchors[1].out = CLIP(y,0.0,1.0);

    x = params->anchors[3].out;
    y = x - duller;
    params->anchors[3].out = CLIP(y,0.0,1.0);
}

void
sharpenGamma _IMPL((params, sharper),
		   GammaParams *params _AND float sharper )
{
    float x, y;

    if ( sharper < 0 || sharper > 1.0 ) return;

    x = params->anchors[1].out;
    y = x - sharper;
    params->anchors[1].out = CLIP(y,0.0,1.0);

    x = params->anchors[3].out;
    y = x + sharper;
    params->anchors[3].out = CLIP(y,0.0,1.0);
}

typedef void (*toColorF) __((XColor*,Color*));
typedef void (*fromColorF) __((Color*,XColor*));
typedef float (*intensF) __((Color*));
typedef void (*setIntensF) __((Color*,float));

static toColorF toColor[] = {
    RGBtoIHS, RGBtoHexIHS, RGBtoYIQ, RGBtoFRGB
  };

static fromColorF fromColor[] = { 
    IHStoRGB, HexIHStoRGB, YIQtoRGB, FRGBtoRGB
  };

static intensF getIntens[] = { 
    IHSintens, HexIHSintens, YIQintens, FRGBintens
  };

static setIntensF setIntens[] = { 
    SetIHSintens, SetHexIHSintens, SetYIQintens, SetFRGBintens
  };

float
gammaCorrect _IMPL((params, in_out),
		   GammaParams *params _AND XColor *in_out )
{
    Color color;
    int model; 
    float intens;
    Anchor *a;
    int i;

    /* NULL params arg means use default. */
    if ( !params ) params = &defaultGammaParams;

    model = (int) params->model;
    (*toColor[model])(in_out,&color);
    intens = (*getIntens[model])(&color);
    for ( i = 0, a = params->anchors; i < 4; ++i, ++a ) {
	if ( intens >= a[0].in && intens <= a[1].in ) {
	    float i0 = a[0].in,  i1 = a[1].in;
	    float o0 = a[0].out, o1 = a[1].out;
	    intens = o0 + (o1-o0)*(intens-i0)/(i1-i0);
	    break;
	}
    }
    (*setIntens[model])(&color,intens);
    (*fromColor[model])(&color,in_out);
	
    return intens;
}

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/AsciiText.h>

#include "ChooseVar.h"

/* 
 * Anchor data entry.
 */

typedef struct {
    Anchor val;
} cvAnchorData;

static void ChangeAnchor __(( Widget w ));
static void InitAnchor __(( cvChoice *p, Boolean new ));
static void ResetAnchor __(( cvChoice *p ));
static Boolean ChangedAnchor __(( cvChoice *p ));

static Widget
CreateAnchor _IMPL((w, p), Widget w _AND cvChoice *p )
{
    Widget form, prompt, in, out;
    cvAnchorData *cv_data = p->cv_data;

    if ( !cv_data )
      cv_data = p->cv_data = XtNew(cvAnchorData);

    if ( !p->widget ) {
	form = p->widget =
	  XtVaCreateManagedWidget("anchor", formWidgetClass, w, NULL);
	prompt = XtVaCreateManagedWidget("prompt", labelWidgetClass, form,
					 XtNlabel, p->prompt,
					 NULL);

	in = cvCreateAsciiTextWidget("in", form);
	XtVaSetValues(in, XtNfromHoriz, prompt, NULL);

	out = cvCreateAsciiTextWidget("out", form);
	XtVaSetValues(out, XtNfromHoriz, in, NULL);
    }

    return p->widget;
}

static void
ChangeAnchor _IMPL((w), Widget w )
{
    cvChoice *p = getCVChoice(w);
    cvAnchorData *cv_data = p->cv_data;
    Anchor *curr = (Anchor *)p->var;
    double atof();
    String str;
    float val;
    char buf[15];

    XtVaGetValues(w, XtNstring, &str, NULL);
    val = atof(str);
    val = CLIP(val,0.0,1.0);
    sprintf(buf, "%g", val);
    XtVaSetValues(w, XtNstring, buf, NULL);

    if ( !strcmp(XtName(w), "in")  ) curr->in = val;
    if ( !strcmp(XtName(w), "out") ) curr->out = val;
}

static void
InitAnchor _IMPL((p, new), cvChoice *p _AND Boolean new )
{
    cvAnchorData *cv_data = p->cv_data;
    Widget in = XtNameToWidget(p->widget, "in");
    Widget out = XtNameToWidget(p->widget, "out");
    Anchor *curr = (Anchor *)p->var;
    char buf[15];

    if ( new ) cv_data->val = *curr;
    if ( !in || !out) {
	fprintf(stderr, "ResetAnchor(): Cannot find \"in/out\" widget.\n");
	return;
    }

    sprintf(buf, "%g", curr->in);
    XtVaSetValues(in, XtNstring, buf, XtNlength, strlen(buf), NULL);
    sprintf(buf, "%g", curr->out);
    XtVaSetValues(out, XtNstring, buf, XtNlength, strlen(buf), NULL);
}

static void
ResetAnchor _IMPL((p), cvChoice *p )
{
    cvAnchorData *cv_data = p->cv_data;
    Widget in = XtNameToWidget(p->widget, "in");
    Widget out = XtNameToWidget(p->widget, "out");
    Anchor *curr = (Anchor *)p->var;
    char buf[15];

    *curr = cv_data->val;
    if ( !in || !out) return;

    sprintf(buf, "%g", curr->in);
    XtVaSetValues(in, XtNstring, buf, XtNlength, strlen(buf), NULL);
    sprintf(buf, "%g", curr->out);
    XtVaSetValues(out, XtNstring, buf, XtNlength, strlen(buf), NULL);
}

static Boolean
ChangedAnchor _IMPL((p), cvChoice *p )
{
    cvAnchorData *cv_data = p->cv_data;
    Anchor *curr = (Anchor *)p->var;

    return  ((curr->in != cv_data->val.in) || 
	     (curr->out != cv_data->val.out));
}

static void 
  cbBrighter __(( Widget w, XtPointer closure, XtPointer call_data ));
static void 
  cbDimmer __(( Widget w, XtPointer closure, XtPointer call_data ));
static void
  cbSharper __(( Widget w, XtPointer closure, XtPointer call_data ));
static void 
  cbDuller __(( Widget w, XtPointer closure, XtPointer call_data ));
static void 
  cbInitDefault __(( Widget w, XtPointer closure, XtPointer call_data ));

static void 
  cbSetDefault __(( Widget w, XtPointer closure, XtPointer call_data ));
static void
  cbSetEnhance __(( Widget w, XtPointer closure, XtPointer call_data ));
static void
  cbSetFlat __(( Widget w, XtPointer closure, XtPointer call_data ));
static void
  cbSetInverse __(( Widget w, XtPointer closure, XtPointer call_data ));
static void
  cbSetDull __(( Widget w, XtPointer closure, XtPointer call_data ));

Boolean
  EditGamma _IMPL((w, params), Widget w _AND GammaParams *params )
{
    Boolean changed;

    Widget cv;
    XtAppContext app_context = XtWidgetToApplicationContext(w);

    cvChoiceDescr Ch[20]; int Cn = 0;

    static XtPointer modelEnumArgs[] = {
	cvEnumPair("IHS", IHS),
	cvEnumPair("HexIHS", HexIHS),
	cvEnumPair("YIQ", YIQ),
	cvEnumPair("RGB", RGB),
	NULL
      };

    cvEnum(Ch[Cn], params->model, 0, "Color model:", NULL, modelEnumArgs);++Cn;
    cvButton(Ch[Cn], "Default", cbSetDefault, "Preset curve:", params); ++Cn;
    cvButton(Ch[Cn], "Enhance", cbSetEnhance, cvbSameLine,     params); ++Cn;
    cvButton(Ch[Cn], "Flat",    cbSetFlat,    cvbSameLine,     params); ++Cn;
    cvButton(Ch[Cn], "Inverse", cbSetInverse, cvbSameLine,     params); ++Cn;
    cvButton(Ch[Cn], "Dull",    cbSetDull,    cvbSameLine,     params); ++Cn;
    cvSpace(Ch[Cn], 5); ++Cn;
    cvGeneric(Ch[Cn], params->anchors[0],Anchor, 0, "  Anchor 1:", NULL, NULL);
    ++Cn;
    cvGeneric(Ch[Cn], params->anchors[1],Anchor, 0, "  Anchor 2:", NULL, NULL);
    ++Cn;
    cvGeneric(Ch[Cn], params->anchors[2],Anchor, 0, "  Anchor 3:", NULL, NULL);
    ++Cn;
    cvGeneric(Ch[Cn], params->anchors[3],Anchor, 0, "  Anchor 4:", NULL, NULL);
    ++Cn;
    cvGeneric(Ch[Cn], params->anchors[4],Anchor, 0, "  Anchor 5:", NULL, NULL);
    ++Cn;
    cvSpace(Ch[Cn], 5); ++Cn;
    cvButton(Ch[Cn], "Brighter", cbBrighter,  "Modify Curve:", params); ++Cn;
    cvButton(Ch[Cn], "Dimmer",   cbDimmer,    cvbSameLine,     params); ++Cn;
    cvButton(Ch[Cn], "Sharper",  cbSharper,   cvbSameLine,     params); ++Cn;
    cvButton(Ch[Cn], "Duller",   cbDuller,    cvbSameLine,     params); ++Cn;
    cvButton(Ch[Cn], "Set Default", cbInitDefault, cvbEndButtons, params);++Cn;
    
    cv = XtVaCreatePopupShell("gammaChoices", chooseVarWidgetClass, w, NULL);
    
    cvRegisterType("Anchor", 0, 
		   CreateAnchor, InitAnchor, ResetAnchor, 
		   ChangeAnchor, ChangedAnchor);

    cvSetChoices(cv, Ch, Cn);
    changed = ChooseVarModal((ChooseVarWidget) cv, XtGrabExclusive, False);
    XtDestroyWidget(cv);
    
    return changed;
}

static void 
  cbBrighter _IMPL((w, closure, call_data),
		   Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    GammaParams *params = (GammaParams *) closure;

    brightenGamma(params, 0.1);
    cvReinitValues(cvw, False);
}

static void 
  cbDimmer _IMPL((w, closure, call_data),
		 Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    GammaParams *params = (GammaParams *) closure;

    dimGamma(params, 0.1);
    cvReinitValues(cvw, False);
}

static void 
  cbSharper _IMPL((w, closure, call_data),
		  Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    GammaParams *params = (GammaParams *) closure;

    sharpenGamma(params, 0.1);
    cvReinitValues(cvw, False);
}

static void 
  cbDuller _IMPL((w, closure, call_data),
		 Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    GammaParams *params = (GammaParams *) closure;

    dullGamma(params, 0.1);
    cvReinitValues(cvw, False);
}

static void 
  cbInitDefault _IMPL((w, closure, call_data),
		      Widget w _AND XtPointer closure _AND 
		      XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    GammaParams *params = (GammaParams *) closure;

    defaultGammaParams = *params;
    cvDone(w, (XtPointer)cvw, call_data);
}

static void 
  cbSetDefault _IMPL((w, closure, call_data),
		     Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    GammaParams *params = (GammaParams *) closure;

    *params = defaultGammaParams;
    cvReinitValues(cvw, False);
}

static void 
  cbSetEnhance _IMPL((w, closure, call_data),
		     Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    GammaParams *params = (GammaParams *) closure;

    *params = enhanceGammaParams;
    cvReinitValues(cvw, False);
}

static void 
  cbSetFlat _IMPL((w, closure, call_data),
		  Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    GammaParams *params = (GammaParams *) closure;

    *params = flatGammaParams;
    cvReinitValues(cvw, False);
}

static void 
  cbSetInverse _IMPL((w, closure, call_data),
		     Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    GammaParams *params = (GammaParams *) closure;

    *params = invGammaParams;
    cvReinitValues(cvw, False);
}

static void 
  cbSetDull _IMPL((w, closure, call_data),
		  Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    GammaParams *params = (GammaParams *) closure;

    *params = dullGammaParams;
    cvReinitValues(cvw, False);
}

