/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: gif_read.c,v 0.9.1.1 1991/11/06 00:09:25 leei Exp $";

/*
 * adapted from code by kirk johnson (tuna@athena.mit.edu).  most of this
 * code is unchanged. -- jim frost 12.31.89
 *
 * gifin.c
 * kirk johnson
 * november 1989
 *
 * routines for reading GIF files
 *
 * Copyright 1989 Kirk L. Johnson (see the included file
 * "kljcpyrght.h" for complete copyright information)
 */

#include <stdlib.h>

#include <vim/proto.h>
#include <vim/pFile.h>

#include <vim/ByteImageP.h>

#include "gif.h"
#include "kljcpyrght.h"

/*
 * gifin return codes
 */
#define GIFIN_SUCCESS       0   /* success */
#define GIFIN_DONE          1   /* no more images */

#define GIFIN_ERR_BAD_SD   -1   /* bad screen descriptor */
#define GIFIN_ERR_BAD_SEP  -2   /* bad image separator */
#define GIFIN_ERR_BAD_SIG  -3   /* bad signature */
#define GIFIN_ERR_EOD      -4   /* unexpected end of raster data */
#define GIFIN_ERR_EOF      -5   /* unexpected end of input stream */
#define GIFIN_ERR_FAO      -6   /* file already open */
#define GIFIN_ERR_IAO      -7   /* image already open */
#define GIFIN_ERR_NFO      -8   /* no file open */
#define GIFIN_ERR_NIO      -9   /* no image open */

/*
 * colormap indices 
 */

#define GIF_RED  0
#define GIF_GRN  1
#define GIF_BLU  2

static int  gifin_load_cmap __((pFILE in, byte cmap[3][256], int ncolors));
static int  gifin_skip_extension __((pFILE in));
static int  gifin_read_data_block __((pFILE in));
static int  gifin_push_string __((int code));
static void gifin_add_string __((int p, int e));
static void gifin_fatal __((char *msg));

/* #defines, typedefs, and such
 */

#define GIF_SIG      "GIF87a"
#define GIF_SIG_LEN  6          /* GIF signature length */
#define GIF_SD_SIZE  7          /* GIF screen descriptor size */
#define GIF_ID_SIZE  9          /* GIF image descriptor size */

#define GIF_SEPARATOR   ','     /* GIF image separator */
#define GIF_EXTENSION   '!'     /* GIF extension block marker */
#define GIF_TERMINATOR  ';'     /* GIF terminator */

#define STAB_SIZE  4096         /* string table size */
#define PSTK_SIZE  1024         /* pixel stack size */

#define NULL_CODE  -1           /* string table null code */

/****
 **
 ** local #defines
 **
 ****/

#define PUSH_PIXEL(p)                                       \
{                                                           \
  if (pstk_idx == PSTK_SIZE)                                \
    gifin_fatal("pixel stack overflow in PUSH_PIXEL()");    \
  else                                                      \
    pstk[pstk_idx++] = (p);                                 \
}


/****
 **
 ** local variables
 **
 ****/

static int interlace_start[4]= { /* start line for interlacing */
  0, 4, 2, 1
};

static int interlace_rate[4]= { /* rate at which we accelerate vertically */
  8, 8, 4, 2
};

static int  root_size;          /* root code size */
static int  clr_code;           /* clear code */
static int  eoi_code;           /* end of information code */
static int  code_size;          /* current code size */
static int  code_mask;          /* current code mask */
static int  prev_code;          /* previous code */

/*
 * NOTE: a long is assumed to be at least 32 bits wide
 */
static long work_data;          /* working bit buffer */
static int  work_bits;          /* working bit count */

static byte buf[256];           /* byte buffer */
static int  buf_cnt;            /* byte count */
static int  buf_idx;            /* buffer index */

static int table_size;          /* string table size */
static int prefix[STAB_SIZE];   /* string table : prefixes */
static int extnsn[STAB_SIZE];   /* string table : extensions */

static byte pstk[PSTK_SIZE];    /* pixel stack */
static int  pstk_idx;           /* pixel stack pointer */


/****
 **
 ** global variables
 **
 ****/

static int  gifin_rast_width;          /* raster width */
static int  gifin_rast_height;         /* raster height */
static byte gifin_g_cmap_flag;         /* global colormap flag */
static int  gifin_g_pixel_bits;        /* bits per pixel, global colormap */
static int  gifin_g_ncolors;           /* number of colors, global colormap */
static byte gifin_g_cmap[3][256];      /* global colormap */
static int  gifin_bg_color;            /* background color index */
static int  gifin_color_bits;          /* bits of color resolution */

static int  gifin_img_left;            /* image position on raster */
static int  gifin_img_top;             /* image position on raster */
static int  gifin_img_width;           /* image width */
static int  gifin_img_height;          /* image height */
static byte gifin_l_cmap_flag;         /* local colormap flag */
static int  gifin_l_pixel_bits;        /* bits per pixel, local colormap */
static int  gifin_l_ncolors;           /* number of colors, local colormap */
static byte gifin_l_cmap[3][256];      /* local colormap */
static byte gifin_interlace_flag;      /* interlace image format flag */

/*
 * open a GIF file, using s as the input stream
 */

static int 
  gifin_open_file _IMPL((in), pFILE in)
{
    /* check GIF signature */
    if (pfread(buf, 1, GIF_SIG_LEN, in) != GIF_SIG_LEN)
      return GIFIN_ERR_EOF;

    buf[GIF_SIG_LEN] = '\0';
    if (strcmp((char *) buf, GIF_SIG) != 0)
      return GIFIN_ERR_BAD_SIG;

    /* read screen descriptor */
    if (pfread(buf, 1, GIF_SD_SIZE, in) != GIF_SD_SIZE)
      return GIFIN_ERR_EOF;

    /* decode screen descriptor */
    gifin_rast_width   = (buf[1] << 8) + buf[0];
    gifin_rast_height  = (buf[3] << 8) + buf[2];
    gifin_g_cmap_flag  = (buf[4] & 0x80) ? 1 : 0;
    gifin_color_bits   = ((buf[4] & 0x70) >> 4) + 1;
    gifin_g_pixel_bits = (buf[4] & 0x07) + 1;
    gifin_bg_color     = buf[5];
    
    if (buf[6] != 0)
      return GIFIN_ERR_BAD_SD;

    /* load global colormap */
    if (gifin_g_cmap_flag) {
	gifin_g_ncolors = (1 << gifin_g_pixel_bits);

	if (gifin_load_cmap(in,gifin_g_cmap, gifin_g_ncolors) != GIFIN_SUCCESS)
	  return GIFIN_ERR_EOF;
    } else {
	gifin_g_ncolors = 0;
    }

    /* done! */
    return GIFIN_SUCCESS;
}


/*
 * open next GIF image in the input stream; returns GIFIN_SUCCESS if
 * successful. if there are no more images, returns GIFIN_DONE. (might
 * also return various GIFIN_ERR codes.)
 */

static int 
  gifin_open_image _IMPL((in), pFILE in)
{
    int i;
    int separator;

    /* skip over any extension blocks */
    do {
	separator = pfgetc(in);
	if (separator == GIF_EXTENSION) {
	    if (gifin_skip_extension(in) != GIFIN_SUCCESS)
	      return GIFIN_ERR_EOF;
	}
    } while (separator == GIF_EXTENSION);

    /* check for end of file marker */
    if (separator == GIF_TERMINATOR)
      return GIFIN_DONE;

    /* make sure we've got an image separator */
    if (separator != GIF_SEPARATOR)
      return GIFIN_ERR_BAD_SEP;

    /* read image descriptor */
    if (pfread(buf, 1, GIF_ID_SIZE, in) != GIF_ID_SIZE)
      return GIFIN_ERR_EOF;

    /* decode image descriptor */
    gifin_img_left       = (buf[1] << 8) + buf[0];
    gifin_img_top        = (buf[3] << 8) + buf[2];
    gifin_img_width      = (buf[5] << 8) + buf[4];
    gifin_img_height     = (buf[7] << 8) + buf[6];
    gifin_l_cmap_flag    = (buf[8] & 0x80) ? 1 : 0;
    gifin_interlace_flag = (buf[8] & 0x40) ? 1 : 0;
    gifin_l_pixel_bits   = (buf[8] & 0x07) + 1;

    /* load local colormap */
    if (gifin_l_cmap_flag) {
	gifin_l_ncolors = (1 << gifin_l_pixel_bits);
	if (gifin_load_cmap(in,gifin_l_cmap, gifin_l_ncolors) != GIFIN_SUCCESS)
	  return GIFIN_ERR_EOF;
    } else {
	gifin_l_ncolors = 0;
    }

    /* initialize raster data stream decoder */
    root_size = pfgetc(in);
    clr_code  = 1 << root_size;
    eoi_code  = clr_code + 1;
    code_size = root_size + 1;
    code_mask = (1 << code_size) - 1;
    work_bits = 0;
    work_data = 0;
    buf_cnt   = 0;
    buf_idx   = 0;

    /* initialize string table */
    for (i=0; i<STAB_SIZE; i++) {
	prefix[i] = NULL_CODE;
	extnsn[i] = i;
    }

    /* initialize pixel stack */
    pstk_idx = 0;

    /* done! */
    return GIFIN_SUCCESS;
}

/*
 * try to read next pixel from the raster, return result in *pel
 */

static int 
  gifin_get_pixel _IMPL((in, pel), pFILE in _AND int *pel)
{
    int  code;
    int  first;
    int  place;
    
    /* decode until there are some pixels on the pixel stack */
    while (pstk_idx == 0) {
	/* load bytes until we have enough bits for another code */
	while (work_bits < code_size) {
	    if (buf_idx == buf_cnt) {
		/* read a new data block */
		if (gifin_read_data_block(in) != GIFIN_SUCCESS)
		  return GIFIN_ERR_EOF;
		
		if (buf_cnt == 0)
		  return GIFIN_ERR_EOD;
	    }
	    
	    work_data |= ((long) buf[buf_idx++]) << work_bits;
	    work_bits += 8;
	}
	
	/* get the next code */
	code        = work_data & code_mask;
	work_data >>= code_size;
	work_bits  -= code_size;
	
	/* interpret the code */
	if (code == clr_code) {
	    /* reset decoder stream */
	    code_size  = root_size + 1;
	    code_mask  = (1 << code_size) - 1;
	    prev_code  = NULL_CODE;
	    table_size = eoi_code + 1;
	} else if (code == eoi_code) {
	    /* Ooops! no more pixels */
	    return GIFIN_ERR_EOF;
	} else if (prev_code == NULL_CODE) {
	    gifin_push_string(code);
	    prev_code = code;
	} else {
	    if (code < table_size) {
		first = gifin_push_string(code);
	    } else {
		place = pstk_idx;
		PUSH_PIXEL(NULL_CODE);
		first = gifin_push_string(prev_code);
		pstk[place] = first;
	    }
	    
	    gifin_add_string(prev_code, first);
	    prev_code = code;
	}
    }
    
    /* pop a pixel off the pixel stack */
    *pel = (int) pstk[--pstk_idx];
    
    /* done! */
    return GIFIN_SUCCESS;
}

/*
 * close an open GIF image
 */

static int 
  gifin_close_image _IMPL((in), pFILE in)
{
    /* skip any remaining raster data */
    do {
	if (gifin_read_data_block(in) != GIFIN_SUCCESS)
	  return GIFIN_ERR_EOF;
    } while (buf_cnt > 0);

    /* done! */
    return GIFIN_SUCCESS;
}

/*
 * load a colormap from the input stream
 */

static int 
  gifin_load_cmap _IMPL((in, cmap, ncolors),
			pFILE in _AND byte cmap[3][256] _AND int ncolors)
{
    int i;
    
    for (i=0; i<ncolors; i++) {
	if (pfread(buf, 1, 3, in) != 3)
	  return GIFIN_ERR_EOF;
	  
	cmap[GIF_RED][i] = buf[GIF_RED];
	cmap[GIF_GRN][i] = buf[GIF_GRN];
	cmap[GIF_BLU][i] = buf[GIF_BLU];
    }
    
    /* done! */
    return GIFIN_SUCCESS;
}
 
/*
 * skip an extension block in the input stream
 */

static int 
  gifin_skip_extension _IMPL((in), pFILE in)
{
    /* get the extension function byte */
    pfgetc(in);
    
    /* skip any remaining raster data */
    do  {
	if (gifin_read_data_block(in) != GIFIN_SUCCESS)
	  return GIFIN_ERR_EOF;
    } while (buf_cnt > 0);
    
    /* done! */
    return GIFIN_SUCCESS;
}

/*
 * read a new data block from the input stream
 */

static int 
  gifin_read_data_block _IMPL((in), pFILE in)
{
    /* read the data block header */
    buf_cnt = pfgetc(in);

    /* read the data block body */
    if (pfread(buf, 1, buf_cnt, in) != buf_cnt)
      return GIFIN_ERR_EOF;

    buf_idx = 0;

    /* done! */
    return GIFIN_SUCCESS;
}

/*
 * push a string (denoted by a code) onto the pixel stack
 * (returns the code of the first pixel in the string)
 */

static int 
  gifin_push_string _IMPL((code), int code)
{
    int rslt;

    while (prefix[code] != NULL_CODE) {
	PUSH_PIXEL(extnsn[code]);
	code = prefix[code];
    }

    PUSH_PIXEL(extnsn[code]);
    rslt = extnsn[code];

    return rslt;
}

/*
 * add a new string to the string table
 */

static void
  gifin_add_string _IMPL((p, e), int p _AND int e)
{
    prefix[table_size] = p;
    extnsn[table_size] = e;

    if ((table_size == code_mask) && (code_size < 12)) {
	code_size += 1;
	code_mask  = (1 << code_size) - 1;
    }

    table_size += 1;
}

/*
 * semi-graceful fatal error mechanism
 */

static void
  gifin_fatal _IMPL((msg), char *msg)
{
    fprintf(stderr,"Error reading GIF file: %s\n", msg);
    exit(0);
}

GenericImage
gif_read _IMPL((in, ident), pFILE in _AND char *ident )
{
    ByteImage im;
    declare_ByteImage(im);
    Arg argv[1]; Cardinal argc;
    int x, y, pixel, pass, scanlen;
    byte *pixptr, *pixline;
    

    if ((gifin_open_file(in) != GIFIN_SUCCESS) || /* read GIF header */
	(gifin_open_image(in) != GIFIN_SUCCESS)) {  /* read image header */
	return(NULL);
    }
    im_width  = gifin_img_width;
    im_height = gifin_img_height;
    im_depth  = gifin_l_cmap_flag? gifin_l_pixel_bits: gifin_g_pixel_bits;

    argc = 0;
    XtSetArg( argv[argc], ImNdepth, im_depth ); ++argc;
    im = (ByteImage) 
      ImCreateImage( ident, byteImageClass, im_width, im_height, argv,argc );
    access_ByteImage(im,im);

    /* if image has a local colormap, override global colormap
     */
    if (gifin_l_cmap_flag) {
	for (x = 0; x < gifin_l_ncolors; x++) {
	    im_cmap->red[x]   = gifin_l_cmap[GIF_RED][x] << 8;
	    im_cmap->green[x] = gifin_l_cmap[GIF_GRN][x] << 8;
	    im_cmap->blue[x]  = gifin_l_cmap[GIF_BLU][x] << 8;
	}
	im_cmap->used = gifin_l_ncolors;
    } else {
	for (x = 0; x < gifin_g_ncolors; x++) {
	    im_cmap->red[x]   = gifin_g_cmap[GIF_RED][x] << 8;
	    im_cmap->green[x] = gifin_g_cmap[GIF_GRN][x] << 8;
	    im_cmap->blue[x]  = gifin_g_cmap[GIF_BLU][x] << 8;
	}
	im_cmap->used = gifin_g_ncolors;
    }
    
    /* interlaced image -- futz with the vertical trace.  i wish i knew what
     * kind of drugs the GIF people were on when they decided that they
     * needed to support interlacing.
     */

    if (gifin_interlace_flag) {
	scanlen = im_height;

	/* interlacing takes four passes to read, each starting at a different
	 * vertical point.
	 */

	for (pass= 0; pass < 4; pass++) {
	    y = interlace_start[pass];
	    scanlen = im_width * interlace_rate[pass];
	    pixline = &im_data[y * im_width];
	    while (y < gifin_img_height) {
		pixptr = pixline;
		for (x = 0; x < gifin_img_width; x++) {
		    if (gifin_get_pixel(in,&pixel) != GIFIN_SUCCESS) {
			fprintf(stderr,"%s: Short read within image data\n", ident);
			y = gifin_img_height; x = gifin_img_width;
		    }
		    *pixptr++ = pixel;
		}
		y += interlace_rate[pass];
		pixline += scanlen;
	    }
	}
    } else {
	/* not an interlaced image, just read in sequentially
	 */

	pixptr= im_data;
	for (y= 0; y < gifin_img_height; y++)
	  for (x= 0; x < gifin_img_width; x++) {
	      if (gifin_get_pixel(in,&pixel) != GIFIN_SUCCESS) {
		  fprintf(stderr,"%s: Short read within image data\n", ident);
		  y = gifin_img_height; 
		  x = gifin_img_width;
	      }
	      *pixptr++ = pixel;
	  }
    }

    end_ByteImage(im,im);
    return (GenericImage)im;
}
