/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: vimage.c,v 0.9.1.8 1991/12/20 12:40:25 leei Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/Form.h>

#include <vim/ImViewport.h>
#include <vim/Image.h>

#include <vim/GenImage.h>
#include <vim/readImage.h>

#include <vim/CMManage.h>

#include <vim/debug.h>
#include <vim/proto.h>
#include <vim/notifier.h>

#include "config.h"

/* This is necessary to force linkage of the local AsciiSrc. */
#ifdef LOCAL_ASCIISRC
# include <X11/Xaw/AsciiSrc.h>
#endif

#ifdef DLD
#include <dld.h>
#endif

/*
 * Data structure for application resources.
 */

typedef struct {
    String visual;
    String dld_files;
    String dld_path;
    Boolean mtrace;
    Boolean mcheck;
    String debug;
} AppData, *AppDataPtr;

AppData app_data;

/* resource list */
static XtResource resources[] = {
    { "dldFiles", "DldFiles", "String", sizeof(String),
	XtOffsetOf(AppData, dld_files), XtRString,
	"-lvimage -lXaw -lXmu -lXt -lXext -lX11"},
    { "dldPath", "DldPath", "String", sizeof(String),
	XtOffsetOf(AppData, dld_path), XtRString, ".:/usr/lib:/lib"},
    { "debug", "Debug", "String", sizeof(String),
	XtOffsetOf(AppData, debug), XtRString, ""},
    { "mtrace", "MTrace", "Boolean", sizeof(Boolean),
	XtOffsetOf(AppData, mtrace), XtRImmediate, (XtPointer) FALSE},
    { "mcheck", "MCheck", "Boolean", sizeof(Boolean),
	XtOffsetOf(AppData, mcheck), XtRImmediate, (XtPointer) TRUE},
    { "visualClass", "VisualClass", "String", sizeof(String),
	XtOffsetOf(AppData, visual), XtRString, 0}
};

/* Command line option list. */
static XrmOptionDescRec options[] = {
    { "-debug",    "*debug",  	     XrmoptionSepArg, NULL },
    { "-mtrace",   "*mtrace",  	     XrmoptionNoArg, "True" },
    { "-nomtrace", "*mtrace",        XrmoptionNoArg, "False" },
    { "-mcheck",   "*mcheck",  	     XrmoptionNoArg, "True" },
    { "-nomcheck", "*mcheck",        XrmoptionNoArg, "False" },
    { "-visual",   "*visualClass",   XrmoptionSepArg, NULL },
    { "-mono",     "*visualClass",   XrmoptionNoArg, "StaticGrey" },
    { "-grey",     "*visualClass",   XrmoptionNoArg, "StaticGrey" },
    { "-gray",     "*visualClass",   XrmoptionNoArg, "StaticGrey" },
    { "-true",     "*visualClass",   XrmoptionNoArg, "TrueColor" }
};

/*
 *	Widget and non-widget resources if the application defaults
 *	file can't be found.
 *	[ Generated automatically from VImage.ad ]
 */
static String fallbackResources[] = {
#include "VImage.ad.h"
	NULL
};

static void actQuit __(( Widget, XEvent*, String*, Cardinal*));

static XtActionsRec actions[] =
{
    /* {name, procedure}, */
    {"vimageQuit",	actQuit}
};

Widget toplevel;
Boolean dld = True;

int im_len = 0;
GenericImage *im;
Widget *vp;

int nvp = 1;

#ifndef _STDLIB_H_
extern int abort();
#endif

int
main _IMPL((argc,argv), Cardinal argc _AND String *argv)
{
    XtAppContext app_context;
    Widget top, imw;
    int ims = 0;
    int nhoriz = 1, nvert = 1;
    int i;

#ifdef LOCAL_ASCIISRC
    WidgetClass _asciiSrc = asciiSrcObjectClass;
#endif

    XukcRegisterConverters();

    top = toplevel = 
      XtAppInitialize(&app_context, "VImage", 
		      options, XtNumber(options),
		      &argc, argv,
		      fallbackResources, 
		      NULL, 0);
    XtGetApplicationResources(toplevel, &app_data, 
			      resources, XtNumber(resources),
			      argv, argc);

    /* Initialize GNU memory checking and tracing. */
    if ( app_data.mcheck ) mcheck(abort);
    if ( app_data.mtrace ) mtrace();

#if DEBUGGING
    /* Initialize debug flags. */
    while ( app_data.debug && *app_data.debug ) {
	static Boolean warned = False;
	switch ( *app_data.debug ) {
	  case 'i': SET_DEBUG(DBG_IM); break;
	  case 'm': SET_DEBUG(DBG_MEM); break;
	  case 'M': SET_DEBUG(DBG_METH); break;
	  case 'd': SET_DEBUG(DBG_DISP); break;
	  case 'c': SET_DEBUG(DBG_CM); break;
	  case 'l': SET_DEBUG(DBG_LD); break;
	  default:
	    if ( !warned )
	      fprintf(stderr, "%s -debug: i=images, m=memory, M=methods, d=display, c=colors\n", 
		      argv[0]);
	    warned = True;
	    break;
	}
	++app_data.debug;
    }
#endif /* DEBUGGING */

    XtAppAddActions( app_context, actions, XtNumber(actions) );
    stderrNotifier( "VImage Notifier" );
    CMInitialize(toplevel, visualClassFromName(app_data.visual));

#ifdef DLD
    {
	int dld_err;
	if ( dld_err = dld_init (dld_find_executable (argv[0]))) {
	    fprintf(stderr, "DLD: Unable to initialize.  Error %d\n", dld_err);
	    dld = False;
	}
	dld_init_load_files(app_data.dld_path, app_data.dld_files);
    }
#endif

    for ( --argc, ++argv; argc; --argc, ++argv ) {
	if ( **argv == '-' ) {
	    if ( !strcmp(*argv,"-dual") ) nvp = 2;
	    else if ( !strcmp(*argv,"-quad") ) nvp = 4;
	    else if ( !strcmp(*argv,"-vps") ) {
		nvp = atoi(*++argv); --argc;
	    } else if ( !strcmp(*argv, "-vpconf") ) {
		sscanf(*++argv, "%dx%d", &nhoriz, &nvert); --argc;
		nvp = nhoriz * nvert;
	    }
	} else {
	    if ( ims >= im_len ) {
		im_len += 16;
		im = (GenericImage *) 
		  (im? 
		   malloc(im_len * sizeof(*im)):
		   realloc(im, im_len * sizeof(*im)));
	    }
	    im[ims] = readImage(*argv); ++ims;
	}
    }

    if ( nvp > 1 )
      top = XtCreateManagedWidget( "top_form", formWidgetClass, toplevel,
				  NULL, 0 );

    if ( nhoriz*nvert != nvp ) {
	nhoriz = (int) ceil(sqrt((double)nvp));
	nvert  = (nvp+nhoriz-1)/nhoriz;
    }

    /* Allocate space for viewports. */
    vp = (Widget *) malloc(nvp * sizeof(*vp));

    for ( i = 0; i < nvp; ++i ) {
	char buf[16];
	sprintf(buf,"viewport%d",i);
	vp[i] = XtVaCreateManagedWidget(buf, imageViewportWidgetClass, top,
		  XtNfromHoriz, (i%nhoriz==0? (Widget)NULL: vp[i-1]),
		  XtNfromVert,  (i/nhoriz==0? (Widget)NULL: vp[i-nhoriz]),
		  XtNleft,   (i%nhoriz==0?        XtChainLeft:   XtRubber),
		  XtNright,  (i%nhoriz==nhoriz-1? XtChainRight:  XtRubber),
		  XtNtop,    (i/nhoriz==0?        XtChainTop:    XtRubber),
		  XtNbottom, (i/nhoriz==nvert-1?  XtChainBottom: XtRubber),
	          NULL );

	sprintf(buf,"image%d",i);
	imw = XtVaCreateManagedWidget(buf,imageWidgetClass,vp[i],NULL);
	
	if ( i < ims && im[i] ) ImPushImageStack((ImageWidget)imw,im[i]);
    }

    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);

    return 0;
}
				
static void 
actQuit _IMPL((w, ev, str, card),
	      Widget w _AND XEvent *ev _AND String *str _AND Cardinal *card )
{
    exit(0);
}
